"""
***************************************************************************
    AlgorithmDialog.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import datetime
import time
from typing import Optional

from qgis.core import (
    Qgis,
    QgsApplication,
    QgsProcessingAlgorithm,
    QgsProcessingAlgRunnerTask,
    QgsProcessingFeatureSourceDefinition,
    QgsProcessingOutputHtml,
    QgsProxyProgressTask,
)
from qgis.gui import (
    QgsGui,
    QgsProcessingAlgorithmWidgetBase,
    QgsProcessingContextGenerator,
    QgsProcessingParametersGenerator,
)
from qgis.PyQt import sip
from qgis.PyQt.QtCore import QCoreApplication
from qgis.PyQt.QtGui import QColor, QPalette
from qgis.PyQt.QtWidgets import (
    QDialogButtonBox,
    QMainWindow,
    QMessageBox,
    QPushButton,
    QWidget,
)
from qgis.utils import iface

from processing.core.exceptions import InvalidOutputExtension, InvalidParameterValue
from processing.core.ProcessingConfig import ProcessingConfig
from processing.core.ProcessingResults import resultsList
from processing.gui.AlgorithmExecutor import execute, execute_in_place, executeIterating
from processing.gui.BatchAlgorithmDialog import BatchAlgorithmDialog
from processing.gui.ParametersPanel import ParametersPanel
from processing.gui.Postprocessing import handleAlgorithmResults
from processing.tools import dataobjects


class AlgorithmWidget(QgsProcessingAlgorithmWidgetBase):
    """
    QgsProcessingAlgorithmWidgetBase with python specific logic
    """

    def __init__(
        self,
        alg: QgsProcessingAlgorithm,
        in_place: bool = False,
        parent: Optional[QMainWindow] = None,
        flags: QgsProcessingAlgorithmWidgetBase.WidgetFlags = QgsProcessingAlgorithmWidgetBase.WidgetFlags(),
        initialState: Qgis.DockableWidgetInitialState = Qgis.DockableWidgetInitialState.RestorePreviousState,
    ):
        super().__init__(
            parent or (iface and iface.mainWindow()),
            flags=flags,
            initialState=initialState,
        )

        self.feedback_dialog = None
        self.in_place = in_place
        self.active_layer = iface.activeLayer() if self.in_place else None

        self.context = None
        self.feedback = None
        self.history_log_id = None
        self.history_details = {}

        self._is_running = False

        self.setAlgorithm(alg)
        self.setMainWidget(self.getParametersPanel(alg, self))

        if not self.in_place:
            self.runAsBatchButton = QPushButton(
                QCoreApplication.translate("AlgorithmDialog", "Run as Batch Process…")
            )
            self.runAsBatchButton.clicked.connect(self.runAsBatch)
            self.buttonBox().addButton(
                self.runAsBatchButton, QDialogButtonBox.ButtonRole.ResetRole
            )  # reset role to ensure left alignment
        else:
            in_place_input_parameter_name = "INPUT"
            if hasattr(alg, "inputParameterName"):
                in_place_input_parameter_name = alg.inputParameterName()

            self.mainWidget().setParameters(
                {in_place_input_parameter_name: self.active_layer}
            )

            self.runAsBatchButton = None

            if self.active_layer:
                self.updateRunButtonText()
                self.active_layer.selectionChanged.connect(self.updateRunButtonText)
                self.active_layer.willBeDeleted.connect(self.close)
                self.setWindowTitle(
                    self.windowTitle() + " | " + self.active_layer.name()
                )

        self.updateRunButtonVisibility()

    def updateRunButtonText(self):
        has_selection = self.active_layer.selectedFeatureCount() > 0
        self.buttonBox().button(QDialogButtonBox.StandardButton.Ok).setText(
            QCoreApplication.translate("AlgorithmDialog", "Modify Selected Features")
            if has_selection
            else QCoreApplication.translate("AlgorithmDialog", "Modify All Features")
        )

    def getParametersPanel(self, alg, parent):
        panel = ParametersPanel(parent, alg, self.in_place, self.active_layer)
        return panel

    def runAsBatch(self):
        self.reject()
        dlg = BatchAlgorithmDialog(self.algorithm().create())
        dlg.exec()

    def resetAdditionalGui(self):
        if not self.in_place:
            self.runAsBatchButton.setEnabled(True)

    def blockAdditionalControlsWhileRunning(self):
        if not self.in_place:
            self.runAsBatchButton.setEnabled(False)

    def setParameters(self, parameters):
        self.mainWidget().setParameters(parameters)

    def flag_invalid_parameter_value(self, message: str, widget: QWidget):
        """
        Highlights a parameter with an invalid value
        """
        try:
            self.buttonBox().accepted.connect(lambda w=widget: w.setPalette(QPalette()))
            palette = widget.palette()
            palette.setColor(QPalette.ColorRole.Base, QColor(255, 255, 0))
            widget.setPalette(palette)
        except:
            pass
        self.messageBar().clearWidgets()
        self.messageBar().pushMessage(
            "",
            self.tr("Wrong or missing parameter value: {0}").format(message),
            level=Qgis.MessageLevel.Warning,
            duration=5,
        )

    def flag_invalid_output_extension(self, message: str, widget: QWidget):
        """
        Highlights a parameter with an invalid output extension
        """
        try:
            self.buttonBox().accepted.connect(lambda w=widget: w.setPalette(QPalette()))
            palette = widget.palette()
            palette.setColor(QPalette.ColorRole.Base, QColor(255, 255, 0))
            widget.setPalette(palette)
        except:
            pass
        self.messageBar().clearWidgets()
        self.messageBar().pushMessage(
            "", message, level=Qgis.MessageLevel.Warning, duration=5
        )

    def createProcessingParameters(
        self, flags=QgsProcessingParametersGenerator.Flags()
    ):
        if self.mainWidget() is None:
            return {}

        try:
            return self.mainWidget().createProcessingParameters(flags)
        except InvalidParameterValue as e:
            self.flag_invalid_parameter_value(e.parameter.description(), e.widget)
        except InvalidOutputExtension as e:
            self.flag_invalid_output_extension(e.message, e.widget)
        return {}

    def processingContext(self):
        if self.context is None:
            self.feedback = self.createFeedback()
            self.context = dataobjects.createContext(self.feedback)

        self.applyContextOverrides(self.context)
        return self.context

    def runAlgorithm(self):
        self.feedback = self.createFeedback()
        self.context = dataobjects.createContext(self.feedback)
        self.applyContextOverrides(self.context)
        self.algorithmAboutToRun.emit(self.context)

        checkCRS = ProcessingConfig.getSetting(ProcessingConfig.WARN_UNMATCHING_CRS)
        try:
            # messy as all heck, but we don't want to call the dialog's implementation of
            # createProcessingParameters as we want to catch the exceptions raised by the
            # parameter panel instead...
            parameters = (
                {}
                if self.mainWidget() is None
                else self.mainWidget().createProcessingParameters()
            )

            if checkCRS and not self.algorithm().validateInputCrs(
                parameters, self.context
            ):
                reply = QMessageBox.question(
                    self,
                    self.tr("Unmatching CRS's"),
                    self.tr(
                        "Parameters do not all use the same CRS. This can "
                        "cause unexpected results.\nDo you want to "
                        "continue?"
                    ),
                    QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                    QMessageBox.StandardButton.No,
                )
                if reply == QMessageBox.StandardButton.No:
                    return
            ok, msg = self.algorithm().checkParameterValues(parameters, self.context)
            if not ok:
                QMessageBox.warning(self, self.tr("Unable to execute algorithm"), msg)
                return

            self.blockControlsWhileRunning()
            self.setExecutedAnyResult(True)
            self.cancelButton().setEnabled(False)

            self.iterateParam = None
            self._is_running = True

            for param in self.algorithm().parameterDefinitions():
                if (
                    isinstance(
                        parameters.get(param.name(), None),
                        QgsProcessingFeatureSourceDefinition,
                    )
                    and parameters[param.name()].flags
                    & QgsProcessingFeatureSourceDefinition.Flag.FlagCreateIndividualOutputPerInputFeature
                ):
                    self.iterateParam = param.name()
                    break

            self.clearProgress()
            self.feedback.pushVersionInfo(self.algorithm().provider())
            if (
                self.algorithm().provider()
                and self.algorithm().provider().warningMessage()
            ):
                self.feedback.reportError(self.algorithm().provider().warningMessage())

            self.feedback.pushInfo(
                QCoreApplication.translate(
                    "AlgorithmDialog", "Algorithm started at: {}"
                ).format(datetime.datetime.now().replace(microsecond=0).isoformat())
            )

            self.setInfo(
                QCoreApplication.translate(
                    "AlgorithmDialog", "<b>Algorithm '{0}' starting&hellip;</b>"
                ).format(self.algorithm().displayName()),
                escapeHtml=False,
            )

            self.feedback.pushInfo(self.tr("Input parameters:"))
            display_params = []
            for k, v in parameters.items():
                display_params.append(
                    "'"
                    + k
                    + "' : "
                    + self.algorithm()
                    .parameterDefinition(k)
                    .valueAsPythonString(v, self.context)
                )
            self.feedback.pushCommandInfo("{ " + ", ".join(display_params) + " }")
            self.feedback.pushInfo("")
            start_time = time.time()

            def elapsed_time(start_time) -> str:
                delta_t = time.time() - start_time
                hours = int(delta_t / 3600)
                minutes = int((delta_t % 3600) / 60)
                seconds = delta_t - hours * 3600 - minutes * 60

                str_hours = [self.tr("hour"), self.tr("hours")][hours > 1]
                str_minutes = [self.tr("minute"), self.tr("minutes")][minutes > 1]
                str_seconds = [self.tr("second"), self.tr("seconds")][seconds != 1]

                if hours > 0:
                    elapsed = f"{delta_t:0.2f} {str_seconds} ({hours} {str_hours} {minutes} {str_minutes} {seconds:0.0f} {str_seconds})"
                elif minutes > 0:
                    elapsed = f"{delta_t:0.2f} {str_seconds} ({minutes} {str_minutes} {seconds:0.0f} {str_seconds})"
                else:
                    elapsed = f"{delta_t:0.2f} {str_seconds}"

                return elapsed

            if self.iterateParam:
                # Make sure the Log tab is visible before executing the algorithm
                try:
                    self.showLog()
                    self.repaint()
                except:
                    pass

                self.cancelButton().setEnabled(
                    self.algorithm().flags() & QgsProcessingAlgorithm.Flag.FlagCanCancel
                )
                if executeIterating(
                    self.algorithm(),
                    parameters,
                    self.iterateParam,
                    self.context,
                    self.feedback,
                ):
                    self.feedback.pushInfo(
                        self.tr("Execution completed in {}").format(
                            elapsed_time(start_time)
                        )
                    )
                    self.cancelButton().setEnabled(False)
                    self.finish(True, parameters, self.context, self.feedback)
                else:
                    self.cancelButton().setEnabled(False)
                    self.resetGui()
            else:
                self.history_details = {
                    "python_command": self.algorithm().asPythonCommand(
                        parameters, self.context
                    ),
                    "algorithm_id": self.algorithm().id(),
                    "parameters": self.algorithm().asMap(parameters, self.context),
                }
                process_command, command_ok = self.algorithm().asQgisProcessCommand(
                    parameters, self.context
                )
                if command_ok:
                    self.history_details["process_command"] = process_command
                self.history_log_id, _ = QgsGui.historyProviderRegistry().addEntry(
                    "processing", self.history_details
                )

                QgsGui.instance().processingRecentAlgorithmLog().push(
                    self.algorithm().id()
                )
                self.cancelButton().setEnabled(
                    self.algorithm().flags() & QgsProcessingAlgorithm.Flag.FlagCanCancel
                )

                def on_complete(ok, results):
                    if sip.isdeleted(self):
                        return

                    if ok:
                        self.feedback.pushInfo(
                            self.tr("Execution completed in {}").format(
                                elapsed_time(start_time)
                            )
                        )
                        self.feedback.pushFormattedResults(
                            self.algorithm(), self.context, results
                        )
                    else:
                        self.feedback.reportError(
                            self.tr("Execution failed after {}").format(
                                elapsed_time(start_time)
                            )
                        )
                    self.feedback.pushInfo("")

                    if self.history_log_id is not None:
                        # can't deepcopy this!
                        self.history_details["results"] = {
                            k: v for k, v in results.items() if k != "CHILD_INPUTS"
                        }
                        self.history_details["log"] = self.feedback.htmlLog()

                        QgsGui.historyProviderRegistry().updateEntry(
                            self.history_log_id, self.history_details
                        )

                    if self.feedback_dialog is not None:
                        self.feedback_dialog.close()
                        self.feedback_dialog.deleteLater()
                        self.feedback_dialog = None

                    self.cancelButton().setEnabled(False)

                    self.finish(
                        ok, results, self.context, self.feedback, in_place=self.in_place
                    )

                    self.feedback = None
                    self.context = None

                if not self.in_place and not (
                    self.algorithm().flags()
                    & QgsProcessingAlgorithm.Flag.FlagNoThreading
                ):
                    # Make sure the Log tab is visible before executing the algorithm
                    self.showLog()

                    task = QgsProcessingAlgRunnerTask(
                        self.algorithm(), parameters, self.context, self.feedback
                    )
                    if task.isCanceled():
                        on_complete(False, {})
                    else:
                        task.executed.connect(on_complete)
                        self.setCurrentTask(task)
                else:
                    self.proxy_progress = QgsProxyProgressTask(
                        QCoreApplication.translate(
                            "AlgorithmDialog", "Executing “{}”"
                        ).format(self.algorithm().displayName())
                    )
                    QgsApplication.taskManager().addTask(self.proxy_progress)
                    self.feedback.progressChanged.connect(
                        self.proxy_progress.setProxyProgress
                    )
                    self.feedback_dialog = self.createProgressDialog()
                    self.feedback_dialog.show()
                    if self.in_place:
                        ok, results = execute_in_place(
                            self.algorithm(), parameters, self.context, self.feedback
                        )
                    else:
                        ok, results = execute(
                            self.algorithm(), parameters, self.context, self.feedback
                        )
                    self.feedback.progressChanged.disconnect()
                    self.proxy_progress.finalize(ok)
                    on_complete(ok, results)

        except InvalidParameterValue as e:
            self.flag_invalid_parameter_value(e.parameter.description(), e.widget)
        except InvalidOutputExtension as e:
            self.flag_invalid_output_extension(e.message, e.widget)

    def finish(self, successful, result, context, feedback, in_place=False):
        keepOpen = not successful or ProcessingConfig.getSetting(
            ProcessingConfig.KEEP_DIALOG_OPEN
        )
        generated_html_outputs = False

        self._is_running = False

        if not in_place and self.iterateParam is None:
            # add html results to results dock
            for out in self.algorithm().outputDefinitions():
                if (
                    isinstance(out, QgsProcessingOutputHtml)
                    and out.name() in result
                    and result[out.name()]
                ):
                    resultsList.addResult(
                        icon=self.algorithm().icon(),
                        name=out.description(),
                        timestamp=time.localtime(),
                        result=result[out.name()],
                    )
                    generated_html_outputs = True
            if not handleAlgorithmResults(self.algorithm(), context, feedback, result):
                self.resetGui()
                return

        self.setExecuted(True)
        self.setResults(result)
        self.setInfo(
            self.tr("Algorithm '{0}' finished").format(self.algorithm().displayName()),
            escapeHtml=False,
        )
        self.algorithmFinished.emit(successful, result)

        if not in_place and not keepOpen:
            self.close()
        else:
            self.resetGui()
            if generated_html_outputs:
                self.setInfo(
                    self.tr(
                        "HTML output has been generated by this algorithm."
                        "\nOpen the results dialog to check it."
                    ),
                    escapeHtml=False,
                )

    def isRunning(self):
        return self._is_running



"""
***************************************************************************
    QgisAlgorithmTests.py
    ---------------------
    Date                 : January 2016
    Copyright            : (C) 2016 by Matthias Kuhn
    Email                : matthias@opengis.ch
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Matthias Kuhn"
__date__ = "January 2016"
__copyright__ = "(C) 2016, Matthias Kuhn"

import os
import shutil
import unittest

import AlgorithmsTestBase
import nose2
from qgis.analysis import QgsNativeAlgorithms
from qgis.core import (
    QgsApplication,
    QgsProcessingAlgorithm,
    QgsProcessingException,
    QgsProcessingFeedback,
)
from qgis.testing import QgisTestCase, start_app

from processing.core.ProcessingConfig import ProcessingConfig
from processing.modeler.ModelerUtils import ModelerUtils
from processing.tools.dataobjects import createContext


class TestAlg(QgsProcessingAlgorithm):
    def __init__(self):
        super().__init__()

    def name(self):
        return "testalg"

    def displayName(self):
        return "testalg"

    def initAlgorithm(self, config=None):
        pass

    def createInstance(self):
        return TestAlg()

    def processAlgorithm(self, parameters, context, feedback):
        raise QgsProcessingException("Exception while processing")
        return {}


class TestQgisAlgorithms(QgisTestCase, AlgorithmsTestBase.AlgorithmsTest):
    @classmethod
    def setUpClass(cls):
        start_app()
        from processing.core.Processing import Processing

        Processing.initialize()
        cls.cleanup_paths = []
        cls.in_place_layers = {}
        cls.vector_layer_params = {}

    @classmethod
    def tearDownClass(cls):
        from processing.core.Processing import Processing

        Processing.deinitialize()
        for path in cls.cleanup_paths:
            shutil.rmtree(path)

    def definition_file(self):
        return "qgis_algorithm_tests1.yaml"

    def testProcessingException(self):
        """
        Test that Python exception is caught when running an alg
        """

        alg = TestAlg()
        context = createContext()
        feedback = QgsProcessingFeedback()
        results, ok = alg.run({}, context, feedback)
        self.assertFalse(ok)

    def testParameterPythonImport(self):
        for t in QgsApplication.processingRegistry().parameterTypes():
            import_string = t.pythonImportString()
            # check that pythonImportString correctly imports
            # and that we can instantiate an object!
            if t.className() == "QgsProcessingParameterProviderConnection":
                exec(
                    f"{import_string}\ntest = {t.className()}('id','name', 'provider')\nassert test is not None",
                    {},
                )
            else:
                (
                    exec(
                        f"{import_string}\ntest = {t.className()}('id','name')\nassert test is not None"
                    ),
                    {},
                )


if __name__ == "__main__":
    nose2.main()



"""
***************************************************************************
    viewshed.py
    ---------------------
    Date                 : October 2019
    Copyright            : (C) 2019 by Alexander Bruy
    Email                : alexander dot bruy at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Alexander Bruy"
__date__ = "October 2019"
__copyright__ = "(C) 2019, Alexander Bruy"

import os

from qgis.core import (
    QgsProcessingException,
    QgsProcessingParameterBand,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterDistance,
    QgsProcessingParameterNumber,
    QgsProcessingParameterPoint,
    QgsProcessingParameterRasterDestination,
    QgsProcessingParameterRasterLayer,
    QgsProcessingParameterString,
    QgsRasterFileWriter,
)

from processing.algs.gdal.GdalAlgorithm import GdalAlgorithm
from processing.algs.gdal.GdalUtils import GdalUtils

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]


class viewshed(GdalAlgorithm):
    INPUT = "INPUT"
    BAND = "BAND"
    OBSERVER = "OBSERVER"
    OBSERVER_HEIGHT = "OBSERVER_HEIGHT"
    TARGET_HEIGHT = "TARGET_HEIGHT"
    MAX_DISTANCE = "MAX_DISTANCE"
    OPTIONS = "OPTIONS"
    CREATION_OPTIONS = "CREATION_OPTIONS"
    EXTRA = "EXTRA"
    OUTPUT = "OUTPUT"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterRasterLayer(self.INPUT, self.tr("Input layer"))
        )
        self.addParameter(
            QgsProcessingParameterBand(
                self.BAND,
                self.tr("Band number"),
                1,
                parentLayerParameterName=self.INPUT,
            )
        )
        self.addParameter(
            QgsProcessingParameterPoint(self.OBSERVER, self.tr("Observer location"))
        )
        self.addParameter(
            QgsProcessingParameterNumber(
                self.OBSERVER_HEIGHT,
                self.tr("Observer height, DEM units"),
                type=QgsProcessingParameterNumber.Type.Double,
                minValue=0.0,
                defaultValue=1.0,
            )
        )
        self.addParameter(
            QgsProcessingParameterNumber(
                self.TARGET_HEIGHT,
                self.tr("Target height, DEM units"),
                type=QgsProcessingParameterNumber.Type.Double,
                minValue=0.0,
                defaultValue=1.0,
            )
        )
        self.addParameter(
            QgsProcessingParameterDistance(
                self.MAX_DISTANCE,
                self.tr("Maximum distance from observer to compute visibility"),
                parentParameterName=self.INPUT,
                minValue=0.0,
                defaultValue=100.0,
            )
        )

        # backwards compatibility parameter
        # TODO QGIS 5: remove parameter and related logic
        options_param = QgsProcessingParameterString(
            self.OPTIONS,
            self.tr("Additional creation options"),
            defaultValue="",
            optional=True,
        )
        options_param.setFlags(
            options_param.flags() | QgsProcessingParameterDefinition.Flag.Hidden
        )
        options_param.setMetadata({"widget_wrapper": {"widget_type": "rasteroptions"}})
        self.addParameter(options_param)

        creation_options_param = QgsProcessingParameterString(
            self.CREATION_OPTIONS,
            self.tr("Additional creation options"),
            defaultValue="",
            optional=True,
        )
        creation_options_param.setFlags(
            creation_options_param.flags()
            | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        creation_options_param.setMetadata(
            {"widget_wrapper": {"widget_type": "rasteroptions"}}
        )
        self.addParameter(creation_options_param)

        extra_param = QgsProcessingParameterString(
            self.EXTRA,
            self.tr("Additional command-line parameters"),
            defaultValue=None,
            optional=True,
        )
        extra_param.setFlags(
            extra_param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        self.addParameter(extra_param)

        self.addParameter(
            QgsProcessingParameterRasterDestination(self.OUTPUT, self.tr("Output"))
        )

    def name(self):
        return "viewshed"

    def displayName(self):
        return self.tr("Viewshed")

    def group(self):
        return self.tr("Raster miscellaneous")

    def groupId(self):
        return "rastermiscellaneous"

    def commandName(self):
        return "gdal_viewshed"

    def getConsoleCommands(self, parameters, context, feedback, executing=True):
        dem = self.parameterAsRasterLayer(parameters, self.INPUT, context)
        if dem is None:
            raise QgsProcessingException(
                self.invalidRasterError(parameters, self.INPUT)
            )
        dem_input_details = GdalUtils.gdal_connection_details_from_layer(dem)

        out = self.parameterAsOutputLayer(parameters, self.OUTPUT, context)
        self.setOutputValue(self.OUTPUT, out)

        observer = self.parameterAsPoint(parameters, self.OBSERVER, context, dem.crs())

        arguments = [
            "-b",
            f"{self.parameterAsInt(parameters, self.BAND, context)}",
            "-ox",
            f"{observer.x()}",
            "-oy",
            f"{observer.y()}",
            "-oz",
            f"{self.parameterAsDouble(parameters, self.OBSERVER_HEIGHT, context)}",
            "-tz",
            f"{self.parameterAsDouble(parameters, self.TARGET_HEIGHT, context)}",
            "-md",
            f"{self.parameterAsDouble(parameters, self.MAX_DISTANCE, context)}",
            "-f",
            self.outputFormat(parameters, self.OUTPUT, context),
        ]

        options = self.parameterAsString(parameters, self.CREATION_OPTIONS, context)
        # handle backwards compatibility parameter OPTIONS
        if self.OPTIONS in parameters and parameters[self.OPTIONS] not in (None, ""):
            options = self.parameterAsString(parameters, self.OPTIONS, context)
        if options:
            arguments.extend(GdalUtils.parseCreationOptions(options))

        if self.EXTRA in parameters and parameters[self.EXTRA] not in (None, ""):
            extra = self.parameterAsString(parameters, self.EXTRA, context)
            arguments.append(extra)

        arguments.append(dem_input_details.connection_string)
        arguments.append(out)

        if dem_input_details.credential_options:
            arguments.extend(dem_input_details.credential_options_as_arguments())

        return [self.commandName(), GdalUtils.escapeAndJoin(arguments)]



"""
***************************************************************************
    r_blend_combine.py
    ------------------
    Date                 : February 2016
    Copyright            : (C) 2016 by Médéric Ribreux
    Email                : medspx at medspx dot fr
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Médéric Ribreux"
__date__ = "February 2016"
__copyright__ = "(C) 2016, Médéric Ribreux"


def processInputs(alg, parameters, context, feedback):
    if "first" and "second" in alg.exportedLayers:
        return

    # Use v.in.ogr
    for name in ["first", "second"]:
        alg.loadRasterLayerFromParameter(name, parameters, context, False, None)
    alg.postInputs(context)


def processOutputs(alg, parameters, context, feedback):
    # Keep color table
    alg.exportRasterLayerFromParameter("output", parameters, context)



"""
***************************************************************************
    sql_dictionary.py
    ---------------------
    Date                 : April 2012
    Copyright            : (C) 2012 by Giuseppe Sucameli
    Email                : brush dot tyler at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""


def getSqlDictionary(spatial=True):
    from ..spatialite.sql_dictionary import getSqlDictionary

    return getSqlDictionary(spatial)


def getQueryBuilderDictionary():
    from ..spatialite.sql_dictionary import getQueryBuilderDictionary

    return getQueryBuilderDictionary()



"""
***************************************************************************
    OpenScriptFromFileAction.py
    ---------------------
    Date                 : May 2018
    Copyright            : (C) 2018 by Mathieu Pellerin
    Email                : nirvn dot asia at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Nyall Dawson"
__date__ = "February 2018"
__copyright__ = "(C) 2018, Nyall Dawson"

import os

from qgis.core import QgsApplication, QgsSettings
from qgis.PyQt.QtCore import QCoreApplication, QFileInfo
from qgis.PyQt.QtWidgets import QFileDialog
from qgis.utils import iface

from processing.gui.ToolboxAction import ToolboxAction
from processing.script.ScriptEditorDialog import ScriptEditorDialog

pluginPath = os.path.split(os.path.dirname(__file__))[0]


class OpenScriptFromFileAction(ToolboxAction):
    def __init__(self):
        self.name = QCoreApplication.translate(
            "OpenScriptFromFileAction", "Open Existing Script…"
        )
        self.group = self.tr("Tools")

    def getIcon(self):
        return QgsApplication.getThemeIcon("/processingScript.svg")

    def execute(self):
        settings = QgsSettings()
        lastDir = settings.value("Processing/lastScriptsDir", "")
        filename, selected_filter = QFileDialog.getOpenFileName(
            self.toolbox,
            self.tr("Open Script", "AddScriptFromFileAction"),
            lastDir,
            self.tr("Processing scripts (*.py *.PY)", "AddScriptFromFileAction"),
        )
        if filename:
            settings.setValue(
                "Processing/lastScriptsDir",
                QFileInfo(filename).absoluteDir().absolutePath(),
            )

            dlg = ScriptEditorDialog(filePath=filename, parent=iface.mainWindow())
            dlg.show()



"""
***************************************************************************
    qgstaskwrapper.py
    ---------------------
    Date                 : May 2018
    Copyright            : (C) 2018 by Denis Rouzaud
    Email                : denis@opengis.ch
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

from qgis._core import QgsTask


class QgsTaskWrapper(QgsTask):
    def __init__(self, description, flags, function, on_finished, *args, **kwargs):
        QgsTask.__init__(self, description, flags)
        self.args = args
        self.kwargs = kwargs
        self.function = function
        self.on_finished = on_finished
        self.returned_values = None
        self.exception = None

    def run(self):
        try:
            self.returned_values = self.function(self, *self.args, **self.kwargs)
        except Exception as ex:
            # report error
            self.exception = ex
            return False

        return True

    def finished(self, result):
        if not self.on_finished:
            return

        if not result and self.exception is None:
            self.exception = Exception("Task canceled")

        try:
            if self.returned_values:
                self.on_finished(self.exception, self.returned_values)
            else:
                self.on_finished(self.exception)
        except Exception as ex:
            self.exception = ex



"""
***************************************************************************
    RectanglesOvalsDiamondsVariable.py
    ---------------------
    Date                 : April 2016
    Copyright            : (C) 2016 by Alexander Bruy
    Email                : alexander dot bruy at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Alexander Bruy"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import math

from qgis.core import (
    NULL,
    QgsFeature,
    QgsFeatureSink,
    QgsGeometry,
    QgsPointXY,
    QgsProcessing,
    QgsProcessingAlgorithm,
    QgsProcessingException,
    QgsProcessingParameterEnum,
    QgsProcessingParameterFeatureSink,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterField,
    QgsProcessingParameterNumber,
    QgsWkbTypes,
)
from qgis.PyQt.QtCore import QCoreApplication

from processing.algs.qgis.QgisAlgorithm import QgisAlgorithm


class RectanglesOvalsDiamondsVariable(QgisAlgorithm):
    INPUT = "INPUT"
    SHAPE = "SHAPE"
    WIDTH = "WIDTH"
    HEIGHT = "HEIGHT"
    ROTATION = "ROTATION"
    SEGMENTS = "SEGMENTS"
    OUTPUT = "OUTPUT"

    def group(self):
        return self.tr("Vector geometry")

    def groupId(self):
        return "vectorgeometry"

    def __init__(self):
        super().__init__()

    def flags(self):
        return super().flags() | QgsProcessingAlgorithm.Flag.FlagDeprecated

    def initAlgorithm(self, config=None):
        self.shapes = [self.tr("Rectangles"), self.tr("Diamonds"), self.tr("Ovals")]

        self.addParameter(
            QgsProcessingParameterFeatureSource(
                self.INPUT,
                self.tr("Input layer"),
                [QgsProcessing.SourceType.TypeVectorPoint],
            )
        )

        self.addParameter(
            QgsProcessingParameterEnum(
                self.SHAPE, self.tr("Buffer shape"), options=self.shapes
            )
        )

        self.addParameter(
            QgsProcessingParameterField(
                self.WIDTH,
                self.tr("Width field"),
                parentLayerParameterName=self.INPUT,
                type=QgsProcessingParameterField.DataType.Numeric,
            )
        )
        self.addParameter(
            QgsProcessingParameterField(
                self.HEIGHT,
                self.tr("Height field"),
                parentLayerParameterName=self.INPUT,
                type=QgsProcessingParameterField.DataType.Numeric,
            )
        )
        self.addParameter(
            QgsProcessingParameterField(
                self.ROTATION,
                self.tr("Rotation field"),
                parentLayerParameterName=self.INPUT,
                type=QgsProcessingParameterField.DataType.Numeric,
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterNumber(
                self.SEGMENTS,
                self.tr("Number of segments"),
                minValue=1,
                defaultValue=36,
            )
        )

        self.addParameter(
            QgsProcessingParameterFeatureSink(
                self.OUTPUT,
                self.tr("Output"),
                type=QgsProcessing.SourceType.TypeVectorPolygon,
            )
        )

    def name(self):
        return "rectanglesovalsdiamondsvariable"

    def displayName(self):
        return self.tr("Rectangles, ovals, diamonds (variable)")

    def shortDescription(self):
        return self.tr(
            "Creates rectangle, oval or diamond-shaped polygons from the input point layer."
        )

    def shortHelpString(self):
        return self.tr(
            "This algorithm creates rectangle, oval or diamond-shaped polygons "
            "from the input point layer using specified width, height and (optional) rotation values. "
            "Multipart inputs should be promoted to singleparts first."
        )

    def processAlgorithm(self, parameters, context, feedback):
        source = self.parameterAsSource(parameters, self.INPUT, context)
        if source is None:
            raise QgsProcessingException(
                self.invalidSourceError(parameters, self.INPUT)
            )

        shape = self.parameterAsEnum(parameters, self.SHAPE, context)

        width_field = self.parameterAsString(parameters, self.WIDTH, context)
        height_field = self.parameterAsString(parameters, self.HEIGHT, context)
        rotation_field = self.parameterAsString(parameters, self.ROTATION, context)
        segments = self.parameterAsInt(parameters, self.SEGMENTS, context)

        (sink, dest_id) = self.parameterAsSink(
            parameters,
            self.OUTPUT,
            context,
            source.fields(),
            QgsWkbTypes.Type.Polygon,
            source.sourceCrs(),
        )
        if sink is None:
            raise QgsProcessingException(self.invalidSinkError(parameters, self.OUTPUT))

        width = source.fields().lookupField(width_field)
        height = source.fields().lookupField(height_field)
        rotation = source.fields().lookupField(rotation_field)
        if shape == 0:
            self.rectangles(sink, source, width, height, rotation, feedback)
        elif shape == 1:
            self.diamonds(sink, source, width, height, rotation, feedback)
        else:
            self.ovals(sink, source, width, height, rotation, segments, feedback)

        sink.finalize()
        feedback.featureSinkFinalized(self.OUTPUT)
        return {self.OUTPUT: dest_id}

    def rectangles(self, sink, source, width, height, rotation, feedback):
        ft = QgsFeature()
        features = source.getFeatures()

        total = 100.0 / source.featureCount() if source.featureCount() else 0

        if rotation >= 0:
            for current, feat in enumerate(features):
                if feedback.isCanceled():
                    break

                if not feat.hasGeometry():
                    continue

                w = feat[width]
                h = feat[height]
                angle = feat[rotation]
                # block 0/NULL width or height, but allow 0 as angle value
                if not w or not h:
                    feedback.pushInfo(
                        QCoreApplication.translate(
                            "RectanglesOvalsDiamondsVariable",
                            "Feature {} has empty width or height. Skipping…",
                        ).format(feat.id())
                    )
                    continue
                if angle is NULL:
                    feedback.pushInfo(
                        QCoreApplication.translate(
                            "RectanglesOvalsDiamondsVariable",
                            "Feature {} has empty angle. Skipping…",
                        ).format(feat.id())
                    )
                    continue

                xOffset = w / 2.0
                yOffset = h / 2.0
                phi = angle * math.pi / 180

                point = feat.geometry().asPoint()
                x = point.x()
                y = point.y()
                points = [
                    (-xOffset, -yOffset),
                    (-xOffset, yOffset),
                    (xOffset, yOffset),
                    (xOffset, -yOffset),
                ]
                polygon = [
                    [
                        QgsPointXY(
                            i[0] * math.cos(phi) + i[1] * math.sin(phi) + x,
                            -i[0] * math.sin(phi) + i[1] * math.cos(phi) + y,
                        )
                        for i in points
                    ]
                ]

                ft.setGeometry(QgsGeometry.fromPolygonXY(polygon))
                ft.setAttributes(feat.attributes())
                sink.addFeature(ft, QgsFeatureSink.Flag.FastInsert)

                feedback.setProgress(int(current * total))
        else:
            for current, feat in enumerate(features):
                if feedback.isCanceled():
                    break

                if not feat.hasGeometry():
                    continue

                w = feat[width]
                h = feat[height]
                if not w or not h:
                    feedback.pushInfo(
                        QCoreApplication.translate(
                            "RectanglesOvalsDiamondsVariable",
                            "Feature {} has empty width or height. Skipping…",
                        ).format(feat.id())
                    )
                    continue

                xOffset = w / 2.0
                yOffset = h / 2.0

                point = feat.geometry().asPoint()
                x = point.x()
                y = point.y()
                points = [
                    (-xOffset, -yOffset),
                    (-xOffset, yOffset),
                    (xOffset, yOffset),
                    (xOffset, -yOffset),
                ]
                polygon = [[QgsPointXY(i[0] + x, i[1] + y) for i in points]]

                ft.setGeometry(QgsGeometry.fromPolygonXY(polygon))
                ft.setAttributes(feat.attributes())
                sink.addFeature(ft, QgsFeatureSink.Flag.FastInsert)

                feedback.setProgress(int(current * total))

    def diamonds(self, sink, source, width, height, rotation, feedback):
        features = source.getFeatures()
        ft = QgsFeature()

        total = 100.0 / source.featureCount() if source.featureCount() else 0
        if rotation >= 0:
            for current, feat in enumerate(features):
                if feedback.isCanceled():
                    break

                if not feat.hasGeometry():
                    continue

                w = feat[width]
                h = feat[height]
                angle = feat[rotation]
                # block 0/NULL width or height, but allow 0 as angle value
                if not w or not h:
                    feedback.pushInfo(
                        QCoreApplication.translate(
                            "RectanglesOvalsDiamondsVariable",
                            "Feature {} has empty width or height. Skipping…",
                        ).format(feat.id())
                    )
                    continue
                if angle is NULL:
                    feedback.pushInfo(
                        QCoreApplication.translate(
                            "RectanglesOvalsDiamondsVariable",
                            "Feature {} has empty angle. Skipping…",
                        ).format(feat.id())
                    )
                    continue

                xOffset = w / 2.0
                yOffset = h / 2.0
                phi = angle * math.pi / 180

                point = feat.geometry().asPoint()
                x = point.x()
                y = point.y()
                points = [
                    (0.0, -yOffset),
                    (-xOffset, 0.0),
                    (0.0, yOffset),
                    (xOffset, 0.0),
                ]
                polygon = [
                    [
                        QgsPointXY(
                            i[0] * math.cos(phi) + i[1] * math.sin(phi) + x,
                            -i[0] * math.sin(phi) + i[1] * math.cos(phi) + y,
                        )
                        for i in points
                    ]
                ]

                ft.setGeometry(QgsGeometry.fromPolygonXY(polygon))
                ft.setAttributes(feat.attributes())
                sink.addFeature(ft, QgsFeatureSink.Flag.FastInsert)
                feedback.setProgress(int(current * total))
        else:
            for current, feat in enumerate(features):
                if feedback.isCanceled():
                    break

                if not feat.hasGeometry():
                    continue

                w = feat[width]
                h = feat[height]
                if not w or not h:
                    feedback.pushInfo(
                        QCoreApplication.translate(
                            "RectanglesOvalsDiamondsVariable",
                            "Feature {} has empty width or height. Skipping…",
                        ).format(feat.id())
                    )
                    continue

                xOffset = w / 2.0
                yOffset = h / 2.0

                point = feat.geometry().asPoint()
                x = point.x()
                y = point.y()
                points = [
                    (0.0, -yOffset),
                    (-xOffset, 0.0),
                    (0.0, yOffset),
                    (xOffset, 0.0),
                ]
                polygon = [[QgsPointXY(i[0] + x, i[1] + y) for i in points]]

                ft.setGeometry(QgsGeometry.fromPolygonXY(polygon))
                ft.setAttributes(feat.attributes())
                sink.addFeature(ft, QgsFeatureSink.Flag.FastInsert)
                feedback.setProgress(int(current * total))

    def ovals(self, sink, source, width, height, rotation, segments, feedback):
        features = source.getFeatures()
        ft = QgsFeature()

        total = 100.0 / source.featureCount() if source.featureCount() else 0
        if rotation >= 0:
            for current, feat in enumerate(features):
                if feedback.isCanceled():
                    break

                if not feat.hasGeometry():
                    continue

                w = feat[width]
                h = feat[height]
                angle = feat[rotation]
                # block 0/NULL width or height, but allow 0 as angle value
                if not w or not h:
                    feedback.pushInfo(
                        QCoreApplication.translate(
                            "RectanglesOvalsDiamondsVariable",
                            "Feature {} has empty width or height. Skipping…",
                        ).format(feat.id())
                    )
                    continue
                if angle == NULL:
                    feedback.pushInfo(
                        QCoreApplication.translate(
                            "RectanglesOvalsDiamondsVariable",
                            "Feature {} has empty angle. Skipping…",
                        ).format(feat.id())
                    )
                    continue

                xOffset = w / 2.0
                yOffset = h / 2.0
                phi = angle * math.pi / 180

                point = feat.geometry().asPoint()
                x = point.x()
                y = point.y()
                points = [
                    (xOffset * math.cos(t), yOffset * math.sin(t))
                    for t in [(2 * math.pi) / segments * i for i in range(segments)]
                ]

                polygon = [
                    [
                        QgsPointXY(
                            i[0] * math.cos(phi) + i[1] * math.sin(phi) + x,
                            -i[0] * math.sin(phi) + i[1] * math.cos(phi) + y,
                        )
                        for i in points
                    ]
                ]

                ft.setGeometry(QgsGeometry.fromPolygonXY(polygon))
                ft.setAttributes(feat.attributes())
                sink.addFeature(ft, QgsFeatureSink.Flag.FastInsert)
                feedback.setProgress(int(current * total))
        else:
            for current, feat in enumerate(features):
                if feedback.isCanceled():
                    break

                if not feat.hasGeometry():
                    continue

                w = feat[width]
                h = feat[height]
                if not w or not h:
                    feedback.pushInfo(
                        QCoreApplication.translate(
                            "RectanglesOvalsDiamondsVariable",
                            "Feature {} has empty width or height. Skipping…",
                        ).format(feat.id())
                    )
                    continue

                xOffset = w / 2.0
                yOffset = h / 2.0

                point = feat.geometry().asPoint()
                x = point.x()
                y = point.y()
                points = [
                    (xOffset * math.cos(t), yOffset * math.sin(t))
                    for t in [(2 * math.pi) / segments * i for i in range(segments)]
                ]

                polygon = [[QgsPointXY(i[0] + x, i[1] + y) for i in points]]

                ft.setGeometry(QgsGeometry.fromPolygonXY(polygon))
                ft.setAttributes(feat.attributes())
                sink.addFeature(ft, QgsFeatureSink.Flag.FastInsert)
                feedback.setProgress(int(current * total))



"""
***************************************************************************
    sql_dictionary.py
    ---------------------
    Date                 : April 2012
    Copyright            : (C) 2012 by Giuseppe Sucameli
    Email                : brush dot tyler at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Giuseppe Sucameli"
__date__ = "April 2012"
__copyright__ = "(C) 2012, Giuseppe Sucameli"

# keywords
keywords = [
    # TODO get them from a reference page
    "action",
    "add",
    "after",
    "all",
    "alter",
    "analyze",
    "and",
    "as",
    "asc",
    "before",
    "begin",
    "between",
    "by",
    "cascade",
    "case",
    "cast",
    "check",
    "collate",
    "column",
    "commit",
    "constraint",
    "create",
    "cross",
    "current_date",
    "current_time",
    "current_timestamp",
    "default",
    "deferrable",
    "deferred",
    "delete",
    "desc",
    "distinct",
    "drop",
    "each",
    "else",
    "end",
    "escape",
    "except",
    "exists",
    "for",
    "foreign",
    "from",
    "full",
    "group",
    "having",
    "ignore",
    "immediate",
    "in",
    "initially",
    "inner",
    "insert",
    "intersect",
    "into",
    "is",
    "isnull",
    "join",
    "key",
    "left",
    "like",
    "limit",
    "match",
    "natural",
    "no",
    "not",
    "notnull",
    "null",
    "of",
    "offset",
    "on",
    "or",
    "order",
    "outer",
    "primary",
    "references",
    "release",
    "restrict",
    "right",
    "rollback",
    "row",
    "savepoint",
    "select",
    "set",
    "table",
    "temporary",
    "then",
    "to",
    "transaction",
    "trigger",
    "union",
    "unique",
    "update",
    "using",
    "values",
    "view",
    "when",
    "where",
    "abort",
    "attach",
    "autoincrement",
    "conflict",
    "database",
    "detach",
    "exclusive",
    "explain",
    "fail",
    "glob",
    "if",
    "index",
    "indexed",
    "instead",
    "plan",
    "pragma",
    "query",
    "raise",
    "regexp",
    "reindex",
    "rename",
    "replace",
    "temp",
    "vacuum",
    "virtual",
]
spatialite_keywords = []

# functions
functions = [
    # TODO get them from a reference page
    "changes",
    "coalesce",
    "glob",
    "ifnull",
    "hex",
    "last_insert_rowid",
    "nullif",
    "quote",
    "random",
    "randomblob",
    "replace",
    "round",
    "soundex",
    "total_change",
    "typeof",
    "zeroblob",
    "date",
    "datetime",
    "julianday",
    "strftime",
]
operators = [
    " AND ",
    " OR ",
    "||",
    " < ",
    " <= ",
    " > ",
    " >= ",
    " = ",
    " <> ",
    " IS ",
    " IS NOT ",
    " IN ",
    " LIKE ",
    " GLOB ",
    " MATCH ",
    " REGEXP ",
]

math_functions = [
    # SQL math functions
    "Abs",
    "ACos",
    "ASin",
    "ATan",
    "Cos",
    "Cot",
    "Degrees",
    "Exp",
    "Floor",
    "Log",
    "Log2",
    "Log10",
    "Pi",
    "Radians",
    "Round",
    "Sign",
    "Sin",
    "Sqrt",
    "StdDev_Pop",
    "StdDev_Samp",
    "Tan",
    "Var_Pop",
    "Var_Samp",
]

string_functions = [
    "Length",
    "Lower",
    "Upper",
    "Like",
    "Trim",
    "LTrim",
    "RTrim",
    "Replace",
    "Substr",
]

aggregate_functions = [
    "Max",
    "Min",
    "Avg",
    "Count",
    "Sum",
    "Group_Concat",
    "Total",
    "Var_Pop",
    "Var_Samp",
    "StdDev_Pop",
    "StdDev_Samp",
]

spatialite_functions = [  # from www.gaia-gis.it/spatialite-2.3.0/spatialite-sql-2.3.0.html
    # SQL utility functions for BLOB objects
    "*iszipblob",
    "*ispdfblob",
    "*isgifblob",
    "*ispngblob",
    "*isjpegblob",
    "*isexifblob",
    "*isexifgpsblob",
    "*geomfromexifgpsblob",
    "MakePoint",
    "BuildMbr",
    "*buildcirclembr",
    "ST_MinX",
    "ST_MinY",
    "ST_MaxX",
    "ST_MaxY",
    # SQL functions for constructing a geometric object given its Well-known Text Representation
    "ST_GeomFromText",
    "*pointfromtext",
    # SQL functions for constructing a geometric object given its Well-known Binary Representation
    "*geomfromwkb",
    "*pointfromwkb",
    # SQL functions for obtaining the Well-known Text / Well-known Binary Representation of a geometric object
    "ST_AsText",
    "ST_AsBinary",
    # SQL functions supporting exotic geometric formats
    "*assvg",
    "*asfgf",
    "*geomfromfgf",
    # SQL functions on type Geometry
    "ST_Dimension",
    "ST_GeometryType",
    "ST_Srid",
    "ST_SetSrid",
    "ST_isEmpty",
    "ST_isSimple",
    "ST_isValid",
    "ST_Boundary",
    "ST_Envelope",
    # SQL functions on type Point
    "ST_X",
    "ST_Y",
    # SQL functions on type Curve [Linestring or Ring]
    "ST_StartPoint",
    "ST_EndPoint",
    "ST_Length",
    "ST_isClosed",
    "ST_isRing",
    "ST_Simplify",
    "*simplifypreservetopology",
    # SQL functions on type LineString
    "ST_NumPoints",
    "ST_PointN",
    # SQL functions on type Surface [Polygon or Ring]
    "ST_Centroid",
    "ST_PointOnSurface",
    "ST_Area",
    # SQL functions on type Polygon
    "ST_ExteriorRing",
    "ST_InteriorRingN",
    # SQL functions on type GeomCollection
    "ST_NumGeometries",
    "ST_GeometryN",
    # SQL functions that test approximative spatial relationships via MBRs
    "MbrEqual",
    "MbrDisjoint",
    "MbrTouches",
    "MbrWithin",
    "MbrOverlaps",
    "MbrIntersects",
    "MbrContains",
    # SQL functions that test spatial relationships
    "ST_Equals",
    "ST_Disjoint",
    "ST_Touches",
    "ST_Within",
    "ST_Overlaps",
    "ST_Crosses",
    "ST_Intersects",
    "ST_Contains",
    "ST_Relate",
    # SQL functions for distance relationships
    "ST_Distance",
    # SQL functions that implement spatial operators
    "ST_Intersection",
    "ST_Difference",
    "ST_Union",
    "ST_SymDifference",
    "ST_Buffer",
    "ST_ConvexHull",
    # SQL functions for coordinate transformations
    "ST_Transform",
    # SQL functions for Spatial-MetaData and Spatial-Index handling
    "*initspatialmetadata",
    "*addgeometrycolumn",
    "*recovergeometrycolumn",
    "*discardgeometrycolumn",
    "*createspatialindex",
    "*creatembrcache",
    "*disablespatialindex",
    # SQL functions implementing FDO/OGR compatibility
    "*checkspatialmetadata",
    "*autofdostart",
    "*autofdostop",
    "*initfdospatialmetadata",
    "*addfdogeometrycolumn",
    "*recoverfdogeometrycolumn",
    "*discardfdogeometrycolumn",
    # SQL functions for MbrCache-based queries
    "*filtermbrwithin",
    "*filtermbrcontains",
    "*filtermbrintersects",
    "*buildmbrfilter",
]

# constants
constants = ["null", "false", "true"]
spatialite_constants = []


def getSqlDictionary(spatial=True):
    def strip_star(s):
        if s[0] == "*":
            return s.lower()[1:]
        else:
            return s.lower()

    k, c, f = list(keywords), list(constants), list(functions)

    if spatial:
        k += spatialite_keywords
        f += spatialite_functions
        c += spatialite_constants

    return {
        "keyword": list(map(strip_star, k)),
        "constant": list(map(strip_star, c)),
        "function": list(map(strip_star, f)),
    }


def getQueryBuilderDictionary():
    # concat functions
    def ff(l):
        return [s for s in l if s[0] != "*"]

    def add_paren(l):
        return [s + "(" for s in l]

    foo = sorted(
        add_paren(ff(list(set.union(set(functions), set(spatialite_functions)))))
    )
    m = sorted(add_paren(ff(math_functions)))
    agg = sorted(add_paren(ff(aggregate_functions)))
    op = ff(operators)
    s = sorted(add_paren(ff(string_functions)))
    return {"function": foo, "math": m, "aggregate": agg, "operator": op, "string": s}



"""
***************************************************************************
    AlgorithmLocatorFilter.py
    -------------------------
    Date                 : May 2017
    Copyright            : (C) 2017 by Nyall Dawson
    Email                : nyall dot dawson at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Nyall Dawson"
__date__ = "May 2017"
__copyright__ = "(C) 2017, Nyall Dawson"

from qgis.core import (
    QgsApplication,
    QgsFields,
    QgsLocatorFilter,
    QgsLocatorResult,
    QgsMapLayerType,
    QgsProcessing,
    QgsProcessingAlgorithm,
    QgsProcessingFeatureBasedAlgorithm,
    QgsStringUtils,
    QgsWkbTypes,
)
from qgis.utils import iface

from processing.core.ProcessingConfig import ProcessingConfig
from processing.gui.algorithm_widget import AlgorithmWidget
from processing.gui.AlgorithmExecutor import execute_in_place
from processing.gui.MessageBarProgress import MessageBarProgress
from processing.gui.MessageDialog import MessageDialog


class AlgorithmLocatorFilter(QgsLocatorFilter):
    def __init__(self, parent=None):
        super().__init__(parent)

    def clone(self):
        return AlgorithmLocatorFilter()

    def name(self):
        return "processing_alg"

    def displayName(self):
        return self.tr("Processing Algorithms")

    def priority(self):
        return QgsLocatorFilter.Priority.Low

    def prefix(self):
        return "a"

    def flags(self):
        return QgsLocatorFilter.Flag.FlagFast

    def fetchResults(self, string, context, feedback):
        # collect results in main thread, since this method is inexpensive and
        # accessing the processing registry is not thread safe
        for a in QgsApplication.processingRegistry().algorithms():
            if a.flags() & QgsProcessingAlgorithm.Flag.FlagHideFromToolbox:
                continue
            if (
                not ProcessingConfig.getSetting(
                    ProcessingConfig.SHOW_ALGORITHMS_KNOWN_ISSUES
                )
                and a.flags() & QgsProcessingAlgorithm.Flag.FlagKnownIssues
            ):
                continue

            result = QgsLocatorResult()
            result.filter = self
            result.displayString = a.displayName()
            result.icon = a.icon()
            result.userData = a.id()
            result.score = 0

            if context.usingPrefix and not string:
                self.resultFetched.emit(result)

            if not string:
                return

            string = string.lower()
            tagScore = 0
            if a.provider():
                tags = [*a.tags(), a.provider().name()]
            else:
                tags = a.tags()

            if a.group():
                tags.append(a.group())

            for t in tags:
                if string in t.lower():
                    tagScore = 1
                    break

            result.score = (
                QgsStringUtils.fuzzyScore(result.displayString, string) * 0.5
                + tagScore * 0.5
            )

            if result.score > 0:
                self.resultFetched.emit(result)

    def triggerResult(self, result):
        alg = QgsApplication.processingRegistry().createAlgorithmById(result.userData)
        if not alg:
            return

        ok, message = alg.canExecute()
        if not ok:
            dlg = MessageDialog()
            dlg.setTitle(self.tr("Missing dependency"))
            dlg.setMessage(message)
            dlg.exec()
            return
        widget = alg.createCustomParametersWidget(parent=iface.mainWindow())
        if not widget:
            widget = AlgorithmWidget(alg)
        widget.exec()


class InPlaceAlgorithmLocatorFilter(QgsLocatorFilter):
    def __init__(self, parent=None):
        super().__init__(parent)

    def clone(self):
        return InPlaceAlgorithmLocatorFilter()

    def name(self):
        return "edit_features"

    def displayName(self):
        return self.tr("Edit Features In-Place")

    def priority(self):
        return QgsLocatorFilter.Priority.Low

    def prefix(self):
        return "ef"

    def flags(self):
        return QgsLocatorFilter.Flag.FlagFast

    def fetchResults(self, string, context, feedback):
        # collect results in main thread, since this method is inexpensive and
        # accessing the processing registry/current layer is not thread safe

        if (
            iface.activeLayer() is None
            or iface.activeLayer().type() != QgsMapLayerType.VectorLayer
        ):
            return

        for a in QgsApplication.processingRegistry().algorithms():
            if not a.flags() & QgsProcessingAlgorithm.Flag.FlagSupportsInPlaceEdits:
                continue

            if not a.supportInPlaceEdit(iface.activeLayer()):
                continue

            result = QgsLocatorResult()
            result.filter = self
            result.displayString = a.displayName()
            result.icon = a.icon()
            result.userData = a.id()
            result.score = 0

            if context.usingPrefix and not string:
                self.resultFetched.emit(result)

            if not string:
                return

            string = string.lower()
            tagScore = 0
            if a.provider():
                tags = [*a.tags(), a.provider().name()]
            else:
                tags = a.tags()

            if a.group():
                tags.append(a.group())

            for t in tags:
                if string in t.lower():
                    tagScore = 1
                    break

            result.score = (
                QgsStringUtils.fuzzyScore(result.displayString, string) * 0.5
                + tagScore * 0.5
            )

            if result.score > 0:
                self.resultFetched.emit(result)

    def triggerResult(self, result):
        config = {"IN_PLACE": True}
        alg = QgsApplication.processingRegistry().createAlgorithmById(
            result.userData, config
        )
        if not alg:
            return

        ok, message = alg.canExecute()
        if not ok:
            dlg = MessageDialog()
            dlg.setTitle(self.tr("Missing dependency"))
            dlg.setMessage(message)
            dlg.exec()
            return

        widget = alg.createCustomParametersWidget(parent=iface.mainWindow())
        if not widget:
            widget = AlgorithmWidget(alg, True)
        widget.exec()



"""
***************************************************************************
    self.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import os

from qgis.core import (
    QgsProcessingException,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterCrs,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterEnum,
    QgsProcessingParameterExtent,
    QgsProcessingParameterNumber,
    QgsProcessingParameterRasterDestination,
    QgsProcessingParameterRasterLayer,
    QgsProcessingParameterString,
    QgsRasterFileWriter,
)
from qgis.PyQt.QtGui import QIcon

from processing.algs.gdal.GdalAlgorithm import GdalAlgorithm
from processing.algs.gdal.GdalUtils import GdalUtils

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]


class warp(GdalAlgorithm):
    INPUT = "INPUT"
    SOURCE_CRS = "SOURCE_CRS"
    TARGET_CRS = "TARGET_CRS"
    NODATA = "NODATA"
    TARGET_RESOLUTION = "TARGET_RESOLUTION"
    OPTIONS = "OPTIONS"
    CREATION_OPTIONS = "CREATION_OPTIONS"
    RESAMPLING = "RESAMPLING"
    DATA_TYPE = "DATA_TYPE"
    TARGET_EXTENT = "TARGET_EXTENT"
    TARGET_EXTENT_CRS = "TARGET_EXTENT_CRS"
    MULTITHREADING = "MULTITHREADING"
    EXTRA = "EXTRA"
    OUTPUT = "OUTPUT"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.methods = (
            (self.tr("Nearest Neighbour"), "near"),
            (self.tr("Bilinear (2x2 Kernel)"), "bilinear"),
            (self.tr("Cubic (4x4 Kernel)"), "cubic"),
            (self.tr("Cubic B-Spline (4x4 Kernel)"), "cubicspline"),
            (self.tr("Lanczos (6x6 Kernel)"), "lanczos"),
            (self.tr("Average"), "average"),
            (self.tr("Mode"), "mode"),
            (self.tr("Maximum"), "max"),
            (self.tr("Minimum"), "min"),
            (self.tr("Median"), "med"),
            (self.tr("First Quartile (Q1)"), "q1"),
            (self.tr("Third Quartile (Q3)"), "q3"),
        )

        self.TYPES = [
            self.tr("Use Input Layer Data Type"),
            "Byte",
            "Int16",
            "UInt16",
            "UInt32",
            "Int32",
            "Float32",
            "Float64",
            "CInt16",
            "CInt32",
            "CFloat32",
            "CFloat64",
            "Int8",
        ]

        self.addParameter(
            QgsProcessingParameterRasterLayer(self.INPUT, self.tr("Input layer"))
        )
        self.addParameter(
            QgsProcessingParameterCrs(
                self.SOURCE_CRS, self.tr("Source CRS"), optional=True
            )
        )
        self.addParameter(
            QgsProcessingParameterCrs(
                self.TARGET_CRS, self.tr("Target CRS"), optional=True
            )
        )
        self.addParameter(
            QgsProcessingParameterEnum(
                self.RESAMPLING,
                self.tr("Resampling method to use"),
                options=[i[0] for i in self.methods],
                defaultValue=0,
            )
        )
        self.addParameter(
            QgsProcessingParameterNumber(
                self.NODATA,
                self.tr("Nodata value for output bands"),
                type=QgsProcessingParameterNumber.Type.Double,
                defaultValue=None,
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterNumber(
                self.TARGET_RESOLUTION,
                self.tr("Output file resolution in target georeferenced units"),
                type=QgsProcessingParameterNumber.Type.Double,
                minValue=0.0,
                defaultValue=None,
                optional=True,
            )
        )

        # backwards compatibility parameter
        # TODO QGIS 5: remove parameter and related logic
        options_param = QgsProcessingParameterString(
            self.OPTIONS,
            self.tr("Additional creation options"),
            defaultValue="",
            optional=True,
        )
        options_param.setFlags(
            options_param.flags() | QgsProcessingParameterDefinition.Flag.Hidden
        )
        options_param.setMetadata({"widget_wrapper": {"widget_type": "rasteroptions"}})
        self.addParameter(options_param)

        creation_options_param = QgsProcessingParameterString(
            self.CREATION_OPTIONS,
            self.tr("Additional creation options"),
            defaultValue="",
            optional=True,
        )
        creation_options_param.setFlags(
            creation_options_param.flags()
            | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        creation_options_param.setMetadata(
            {"widget_wrapper": {"widget_type": "rasteroptions"}}
        )
        self.addParameter(creation_options_param)

        dataType_param = QgsProcessingParameterEnum(
            self.DATA_TYPE,
            self.tr("Output data type"),
            self.TYPES,
            allowMultiple=False,
            defaultValue=0,
        )
        dataType_param.setFlags(
            dataType_param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        self.addParameter(dataType_param)

        target_extent_param = QgsProcessingParameterExtent(
            self.TARGET_EXTENT,
            self.tr("Georeferenced extents of output file to be created"),
            optional=True,
        )
        target_extent_param.setFlags(
            target_extent_param.flags()
            | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        self.addParameter(target_extent_param)

        target_extent_crs_param = QgsProcessingParameterCrs(
            self.TARGET_EXTENT_CRS,
            self.tr("CRS of the target raster extent"),
            optional=True,
        )
        target_extent_crs_param.setFlags(
            target_extent_crs_param.flags()
            | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        self.addParameter(target_extent_crs_param)

        multithreading_param = QgsProcessingParameterBoolean(
            self.MULTITHREADING,
            self.tr("Use multithreaded warping implementation"),
            defaultValue=False,
        )
        multithreading_param.setFlags(
            multithreading_param.flags()
            | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        self.addParameter(multithreading_param)

        extra_param = QgsProcessingParameterString(
            self.EXTRA,
            self.tr("Additional command-line parameters"),
            defaultValue=None,
            optional=True,
        )
        extra_param.setFlags(
            extra_param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        self.addParameter(extra_param)

        self.addParameter(
            QgsProcessingParameterRasterDestination(self.OUTPUT, self.tr("Reprojected"))
        )

    def name(self):
        return "warpreproject"

    def displayName(self):
        return self.tr("Warp (reproject)")

    def group(self):
        return self.tr("Raster projections")

    def groupId(self):
        return "rasterprojections"

    def icon(self):
        return QIcon(os.path.join(pluginPath, "images", "gdaltools", "warp.png"))

    def commandName(self):
        return "gdalwarp"

    def tags(self):
        tags = self.tr("transform,reproject,crs,srs,resample").split(",")
        tags.extend(super().tags())
        return tags

    def getConsoleCommands(self, parameters, context, feedback, executing=True):
        inLayer = self.parameterAsRasterLayer(parameters, self.INPUT, context)
        if inLayer is None:
            raise QgsProcessingException(
                self.invalidRasterError(parameters, self.INPUT)
            )
        input_details = GdalUtils.gdal_connection_details_from_layer(inLayer)

        sourceCrs = self.parameterAsCrs(parameters, self.SOURCE_CRS, context)
        targetCrs = self.parameterAsCrs(parameters, self.TARGET_CRS, context)
        if self.NODATA in parameters and parameters[self.NODATA] is not None:
            nodata = self.parameterAsDouble(parameters, self.NODATA, context)
        else:
            nodata = None
        resolution = self.parameterAsDouble(parameters, self.TARGET_RESOLUTION, context)

        arguments = ["-overwrite"]
        if sourceCrs.isValid():
            arguments.append("-s_srs")
            arguments.append(GdalUtils.gdal_crs_string(sourceCrs))

        if targetCrs.isValid():
            arguments.append("-t_srs")
            arguments.append(GdalUtils.gdal_crs_string(targetCrs))

        if nodata is not None:
            arguments.append("-dstnodata")
            arguments.append(str(nodata))

        if resolution:
            arguments.append("-tr")
            arguments.append(str(resolution))
            arguments.append(str(resolution))

        arguments.append("-r")
        arguments.append(
            self.methods[self.parameterAsEnum(parameters, self.RESAMPLING, context)][1]
        )

        extent = self.parameterAsExtent(parameters, self.TARGET_EXTENT, context)
        if not extent.isNull():
            arguments.append("-te")
            arguments.append(extent.xMinimum())
            arguments.append(extent.yMinimum())
            arguments.append(extent.xMaximum())
            arguments.append(extent.yMaximum())

            extentCrs = self.parameterAsCrs(parameters, self.TARGET_EXTENT_CRS, context)
            if extentCrs.isValid():
                arguments.append("-te_srs")
                arguments.append(GdalUtils.gdal_crs_string(extentCrs))

        if self.parameterAsBoolean(parameters, self.MULTITHREADING, context):
            arguments.append("-multi")

        data_type = self.parameterAsEnum(parameters, self.DATA_TYPE, context)
        if data_type:
            if self.TYPES[data_type] == "Int8" and GdalUtils.version() < 3070000:
                raise QgsProcessingException(
                    self.tr("Int8 data type requires GDAL version 3.7 or later")
                )

            arguments.append("-ot " + self.TYPES[data_type])

        out = self.parameterAsOutputLayer(parameters, self.OUTPUT, context)
        self.setOutputValue(self.OUTPUT, out)

        output_format = self.outputFormat(parameters, self.OUTPUT, context)
        if not output_format:
            raise QgsProcessingException(self.tr("Output format is invalid"))

        arguments.append("-of")
        arguments.append(output_format)

        options = self.parameterAsString(parameters, self.CREATION_OPTIONS, context)
        # handle backwards compatibility parameter OPTIONS
        if self.OPTIONS in parameters and parameters[self.OPTIONS] not in (None, ""):
            options = self.parameterAsString(parameters, self.OPTIONS, context)
        if options:
            arguments.extend(GdalUtils.parseCreationOptions(options))

        if self.EXTRA in parameters and parameters[self.EXTRA] not in (None, ""):
            extra = self.parameterAsString(parameters, self.EXTRA, context)
            arguments.append(extra)

        arguments.append(input_details.connection_string)
        arguments.append(out)

        if input_details.open_options:
            arguments.extend(input_details.open_options_as_arguments())

        if input_details.credential_options:
            arguments.extend(input_details.credential_options_as_arguments())

        return [self.commandName(), GdalUtils.escapeAndJoin(arguments)]



"""
***************************************************************************
    Parameters.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import sys

from qgis.core import (
    Qgis,
    QgsCoordinateReferenceSystem,
    QgsExpression,
    QgsMapLayer,
    QgsProcessing,
    QgsProcessingParameterAnnotationLayer,
    QgsProcessingParameterBand,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterColor,
    QgsProcessingParameterCrs,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterEnum,
    QgsProcessingParameterExpression,
    QgsProcessingParameterExtent,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterField,
    QgsProcessingParameterFile,
    QgsProcessingParameterFileDestination,
    QgsProcessingParameterFolderDestination,
    QgsProcessingParameterGeometry,
    QgsProcessingParameterMapLayer,
    QgsProcessingParameterMatrix,
    QgsProcessingParameterMultipleLayers,
    QgsProcessingParameterNumber,
    QgsProcessingParameterPoint,
    QgsProcessingParameterPointCloudDestination,
    QgsProcessingParameterPointCloudLayer,
    QgsProcessingParameterRange,
    QgsProcessingParameterRasterDestination,
    QgsProcessingParameterRasterLayer,
    QgsProcessingParameters,
    QgsProcessingParameterString,
    QgsProcessingParameterVectorDestination,
    QgsProcessingParameterVectorLayer,
    QgsProcessingUtils,
    QgsProject,
    QgsRasterLayer,
    QgsRectangle,
    QgsVectorFileWriter,
    QgsVectorLayer,
    QgsWkbTypes,
)
from qgis.PyQt.QtCore import QCoreApplication

PARAMETER_NUMBER = "number"
PARAMETER_DISTANCE = "distance"
PARAMETER_SCALE = "scale"
PARAMETER_RASTER = "raster"
PARAMETER_TABLE = "vector"
PARAMETER_VECTOR = "source"
PARAMETER_STRING = "string"
PARAMETER_EXPRESSION = "expression"
PARAMETER_BOOLEAN = "boolean"
PARAMETER_TABLE_FIELD = "field"
PARAMETER_EXTENT = "extent"
PARAMETER_FILE = "file"
PARAMETER_POINT = "point"
PARAMETER_GEOMETRY = "geometry"
PARAMETER_CRS = "crs"
PARAMETER_MULTIPLE = "multilayer"
PARAMETER_BAND = "band"
PARAMETER_LAYOUTITEM = "layoutitem"
PARAMETER_MAP_LAYER = "layer"
PARAMETER_RANGE = "range"
PARAMETER_ENUM = "enum"
PARAMETER_MATRIX = "matrix"
PARAMETER_VECTOR_DESTINATION = "vectorDestination"
PARAMETER_FILE_DESTINATION = "fileDestination"
PARAMETER_FOLDER_DESTINATION = "folderDestination"
PARAMETER_RASTER_DESTINATION = "rasterDestination"
PARAMETER_POINTCLOUD_DESTINATION = "pointCloudDestination"


def getParameterFromString(s, context=""):
    # Try the parameter definitions used in description files
    if "|" in s and (
        s.startswith("QgsProcessingParameter")
        or s.startswith("*QgsProcessingParameter")
        or s.startswith("Parameter")
        or s.startswith("*Parameter")
    ):
        isAdvanced = False
        if s.startswith("*"):
            s = s[1:]
            isAdvanced = True
        tokens = s.split("|")
        params = [t if str(t) != str(None) else None for t in tokens[1:]]

        if True:
            clazz = getattr(sys.modules[__name__], tokens[0])
            # convert to correct type
            if clazz == QgsProcessingParameterRasterLayer:
                if len(params) > 3:
                    params[3] = True if params[3].lower() == "true" else False
            elif clazz == QgsProcessingParameterPointCloudLayer:
                if len(params) > 3:
                    params[3] = True if params[3].lower() == "true" else False
            elif clazz == QgsProcessingParameterAnnotationLayer:
                if len(params) > 3:
                    params[3] = True if params[3].lower() == "true" else False
            elif clazz == QgsProcessingParameterBand:
                if len(params) > 4:
                    params[4] = True if params[4].lower() == "true" else False
                if len(params) > 5:
                    params[5] = True if params[5].lower() == "true" else False
            elif clazz == QgsProcessingParameterVectorLayer:
                if len(params) > 2:
                    try:
                        params[2] = [int(p) for p in params[2].split(";")]
                    except ValueError:
                        params[2] = [
                            getattr(QgsProcessing, p.split(".")[1])
                            for p in params[2].split(";")
                        ]
                if len(params) > 4:
                    params[4] = True if params[4].lower() == "true" else False
            elif clazz == QgsProcessingParameterMapLayer:
                if len(params) > 3:
                    params[3] = True if params[3].lower() == "true" else False
                    try:
                        params[4] = [int(p) for p in params[4].split(";")]
                    except ValueError:
                        params[4] = [
                            getattr(QgsProcessing, p.split(".")[1])
                            for p in params[4].split(";")
                        ]
            elif clazz == QgsProcessingParameterBoolean:
                if len(params) > 2:
                    params[2] = True if params[2].lower() == "true" else False
                if len(params) > 3:
                    params[3] = True if params[3].lower() == "true" else False
            elif clazz == QgsProcessingParameterPoint:
                if len(params) > 3:
                    params[3] = True if params[3].lower() == "true" else False
            elif clazz == QgsProcessingParameterGeometry:
                if len(params) > 3:
                    params[3] = True if params[3].lower() == "true" else False
                if len(params) > 4:
                    try:
                        params[4] = [int(p) for p in params[4].split(";")]
                    except ValueError:
                        params[4] = [
                            getattr(QgsWkbTypes, p.split(".")[1])
                            for p in params[4].split(";")
                        ]
                if len(params) > 5:
                    params[5] = True if params[5].lower() == "true" else False
            elif clazz == QgsProcessingParameterCrs:
                if len(params) > 3:
                    params[3] = True if params[3].lower() == "true" else False
            elif clazz == QgsProcessingParameterRange:
                if len(params) > 2:
                    try:
                        params[2] = Qgis.ProcessingNumberParameterType(int(params[2]))
                    except ValueError:
                        params[2] = getattr(
                            QgsProcessingParameterNumber, params[2].split(".")[1]
                        )
                if len(params) > 4:
                    params[4] = True if params[4].lower() == "true" else False
            elif clazz == QgsProcessingParameterExtent:
                if len(params) > 3:
                    params[3] = True if params[3].lower() == "true" else False
            elif clazz == QgsProcessingParameterExpression:
                if len(params) > 3:
                    params[3] = True if params[3].lower() == "true" else False
            elif clazz == QgsProcessingParameterEnum:
                if len(params) > 2:
                    params[2] = params[2].split(";")
                if len(params) > 3:
                    params[3] = True if params[3].lower() == "true" else False
                if len(params) > 4:
                    # For multiple values; default value is a list of int
                    if params[3] is True:
                        params[4] = (
                            [int(v) for v in params[4].split(",")]
                            if params[4] is not None
                            else None
                        )
                    else:
                        params[4] = int(params[4]) if params[4] is not None else None
                if len(params) > 5:
                    params[5] = True if params[5].lower() == "true" else False
            elif clazz == QgsProcessingParameterFeatureSource:
                if len(params) > 2:
                    try:
                        params[2] = [int(p) for p in params[2].split(";")]
                    except ValueError:
                        params[2] = [
                            getattr(QgsProcessing, p.split(".")[1])
                            for p in params[2].split(";")
                        ]
                if len(params) > 4:
                    params[4] = True if params[4].lower() == "true" else False
            elif clazz == QgsProcessingParameterMultipleLayers:
                if len(params) > 2:
                    try:
                        params[2] = Qgis.ProcessingSourceType(int(params[2]))
                    except ValueError:
                        params[2] = getattr(QgsProcessing, params[2].split(".")[1])
                if len(params) > 4:
                    params[4] = True if params[4].lower() == "true" else False
            elif clazz == QgsProcessingParameterMatrix:
                if len(params) > 2:
                    params[2] = int(params[2])
                if len(params) > 3:
                    params[3] = True if params[3].lower() == "true" else False
                if len(params) > 4:
                    params[4] = params[4].split(";")
                if len(params) > 6:
                    params[6] = True if params[6].lower() == "true" else False
            elif clazz == QgsProcessingParameterField:
                if len(params) > 4:
                    try:
                        params[4] = Qgis.ProcessingFieldParameterDataType(
                            int(params[4])
                        )
                    except ValueError:
                        params[4] = getattr(
                            QgsProcessingParameterField, params[4].split(".")[1]
                        )
                if len(params) > 5:
                    params[5] = True if params[5].lower() == "true" else False
                if len(params) > 6:
                    params[6] = True if params[6].lower() == "true" else False
                if len(params) > 7:
                    params[7] = True if params[7].lower() == "true" else False
            elif clazz == QgsProcessingParameterFile:
                if len(params) > 2:
                    try:
                        params[2] = Qgis.ProcessingFileParameterBehavior(int(params[2]))
                    except ValueError:
                        params[2] = getattr(
                            QgsProcessingParameterFile, params[2].split(".")[1]
                        )
                if len(params) > 5:
                    params[5] = True if params[5].lower() == "true" else False
            elif clazz == QgsProcessingParameterNumber:
                if len(params) > 2:
                    try:
                        params[2] = Qgis.ProcessingNumberParameterType(int(params[2]))
                    except ValueError:
                        params[2] = getattr(
                            QgsProcessingParameterNumber, params[2].split(".")[1]
                        )
                if len(params) > 3:
                    params[3] = (
                        float(params[3].strip()) if params[3] is not None else None
                    )
                if len(params) > 4:
                    params[4] = True if params[4].lower() == "true" else False
                if len(params) > 5:
                    params[5] = (
                        float(params[5].strip())
                        if params[5] is not None
                        else -sys.float_info.max + 1
                    )
                if len(params) > 6:
                    params[6] = (
                        float(params[6].strip())
                        if params[6] is not None
                        else sys.float_info.max - 1
                    )
            elif clazz == QgsProcessingParameterString:
                if len(params) > 3:
                    params[3] = True if params[3].lower() == "true" else False
                if len(params) > 4:
                    params[4] = True if params[4].lower() == "true" else False
            elif clazz == QgsProcessingParameterColor:
                if len(params) > 3:
                    params[3] = True if params[3].lower() == "true" else False
                if len(params) > 4:
                    params[4] = True if params[4].lower() == "true" else False
            elif clazz == QgsProcessingParameterFileDestination:
                if len(params) > 4:
                    params[4] = True if params[4].lower() == "true" else False
                if len(params) > 5:
                    params[5] = True if params[5].lower() == "true" else False
            elif clazz == QgsProcessingParameterFolderDestination:
                if len(params) > 3:
                    params[3] = True if params[3].lower() == "true" else False
                if len(params) > 4:
                    params[4] = True if params[4].lower() == "true" else False
            elif clazz == QgsProcessingParameterRasterDestination:
                if len(params) > 3:
                    params[3] = True if params[3].lower() == "true" else False
                if len(params) > 4:
                    params[4] = True if params[4].lower() == "true" else False
            elif clazz == QgsProcessingParameterPointCloudDestination:
                print(params)
                if len(params) > 3:
                    params[3] = True if params[3].lower() == "true" else False
                if len(params) > 4:
                    params[4] = True if params[4].lower() == "true" else False
            elif clazz == QgsProcessingParameterVectorDestination:
                if len(params) > 2:
                    try:
                        params[2] = Qgis.ProcessingSourceType(int(params[2]))
                    except ValueError:
                        params[2] = getattr(QgsProcessing, params[2].split(".")[1])
                if len(params) > 4:
                    params[4] = True if params[4].lower() == "true" else False
                if len(params) > 5:
                    params[5] = True if params[5].lower() == "true" else False

            param = clazz(*params)
            if isAdvanced:
                param.setFlags(
                    param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
                )

            param.setDescription(
                QCoreApplication.translate(context, param.description())
            )

            return param
        else:
            return None
    else:  # try script syntax
        # try native method
        param = QgsProcessingParameters.parameterFromScriptCode(s)
        if param:
            return param



"""
***************************************************************************
    TestData.py
    ---------------------
    Date                 : March 2013
    Copyright            : (C) 2013 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "March 2013"
__copyright__ = "(C) 2013, Victor Olaya"

import os.path
import tempfile

from qgis.core import QgsRasterLayer
from qgis.testing import QgisTestCase

testDataPath = os.path.join(os.path.dirname(__file__), "testdata")


def table():
    return os.path.join(testDataPath, "table.dbf")


def points():
    return os.path.join(testDataPath, "points.gml")


def invalid_geometries():
    return os.path.join(testDataPath, "invalidgeometries.gml")


def wms_layer_1_1_1(test_id: str, crs: str) -> QgsRasterLayer:
    basetestpath = tempfile.mkdtemp().replace("\\", "/")

    endpoint = basetestpath + f"/{test_id}_fake_qgis_http_endpoint"
    with open(
        QgisTestCase.sanitize_local_url(
            endpoint, "?SERVICE=WMS&REQUEST=GetCapabilities"
        ),
        "wb",
    ) as f:
        f.write(
            b"""
            <WMT_MS_Capabilities version="1.1.1">
            <Service>
              <Name>OGC:WMS</Name>
              <Title>OpenStreetMap WMS</Title>
              <Abstract>OpenStreetMap WMS, bereitgestellt durch terrestris GmbH und Co. KG. Beschleunigt mit MapProxy (http://mapproxy.org/)</Abstract>
              <OnlineResource xmlns:xlink="http://www.w3.org/1999/xlink" xlink:href="http://www.terrestris.de"/>
              <AccessConstraints>(c) OpenStreetMap contributors (http://www.openstreetmap.org/copyright) (c) OpenStreetMap Data (http://openstreetmapdata.com) (c) Natural Earth Data (http://www.naturalearthdata.com) (c) GEBCO Compilation Group (2021) GEBCO 2021 Grid (doi:10.5285/c6612cbe-50b3-0cff-e053-6c86abc09f8f) (c) SRTM 450m by ViewfinderPanoramas (http://viewfinderpanoramas.org/) (c) Great Lakes Bathymetry by NGDC (http://www.ngdc.noaa.gov/mgg/greatlakes/) (c) SRTM 30m by NASA EOSDIS Land Processes Distributed Active Archive Center (LP DAAC, https://lpdaac.usgs.gov/) by using this service you agree to the privacy policy mentioned at https://www.terrestris.de/en/datenschutzerklaerung/</AccessConstraints>
            </Service>
            <Capability>
              <Request>
                <GetCapabilities>
                  <Format>application/vnd.ogc.wms_xml</Format>
                </GetCapabilities>
                <GetMap>
                    <Format>image/jpeg</Format>
                    <Format>image/png</Format>
                </GetMap>
                <GetFeatureInfo>
                  <Format>text/plain</Format>
                  <Format>text/html</Format>
                  <Format>application/vnd.ogc.gml</Format>
                </GetFeatureInfo>
                <GetLegendGraphic>
                    <Format>image/jpeg</Format>
                    <Format>image/png</Format>
                    <DCPType>
                        <HTTP>
                            <Get><OnlineResource xmlns:xlink="http://www.w3.org/1999/xlink" xlink:href="https://ows.terrestris.de/osm/service?"/></Get>
                        </HTTP>
                    </DCPType>
                </GetLegendGraphic>
              </Request>
              <Exception>
                <Format>application/vnd.ogc.se_xml</Format>
                <Format>application/vnd.ogc.se_inimage</Format>
                <Format>application/vnd.ogc.se_blank</Format>
              </Exception>
              <Layer queryable="1">
                <Title>OpenStreetMap WMS</Title>
                <SRS>EPSG:3857</SRS>
                <SRS>EPSG:4326</SRS>
                <SRS>EPSG:4686</SRS>
                <LatLonBoundingBox minx="-180" miny="-89.999999" maxx="180" maxy="89.999999" />
                <BoundingBox SRS="EPSG:900913" minx="-20037508.3428" miny="-147730762.67" maxx="20037508.3428" maxy="147730758.195" />
                <BoundingBox SRS="EPSG:4326" minx="-180.0" miny="-90.0" maxx="180.0" maxy="90.0" />
                <BoundingBox SRS="EPSG:3857" minx="-20037508.3428" miny="-147730762.67" maxx="20037508.3428" maxy="147730758.195" />
                <Layer queryable="1">
                  <Name>OSM-WMS</Name>
                  <Title>OpenStreetMap WMS - by terrestris</Title>
                  <LatLonBoundingBox minx="-180" miny="-88" maxx="180" maxy="88" />
                  <BoundingBox SRS="EPSG:900913" minx="-20037508.3428" miny="-25819498.5135" maxx="20037508.3428" maxy="25819498.5135" />
                  <BoundingBox SRS="EPSG:4326" minx="-180" miny="-88" maxx="180" maxy="88" />
                  <BoundingBox SRS="EPSG:3857" minx="-20037508.3428" miny="-25819498.5135" maxx="20037508.3428" maxy="25819498.5135" />
                  <Style>
                      <Name>default</Name>
                      <Title>default</Title>
                      <LegendURL width="155" height="344">
                          <Format>image/png</Format>
                          <OnlineResource xmlns:xlink="http://www.w3.org/1999/xlink" xlink:type="simple" xlink:href="https://ows.terrestris.de/osm/service?styles=&amp;layer=OSM-WMS&amp;service=WMS&amp;format=image%2Fpng&amp;sld_version=1.1.0&amp;request=GetLegendGraphic&amp;version=1.1.1"/>
                      </LegendURL>
                  </Style>
                </Layer>
              </Layer>
            </Capability>
            </WMT_MS_Capabilities>"""
        )
    return QgsRasterLayer(
        f"crs={crs}&layers=OSM-WMS&styles&url=file://"
        + QgisTestCase.sanitize_local_url(
            endpoint, "?SERVICE=WMS&REQUEST=GetCapabilities"
        ),
        "test",
        "wms",
    )


def wms_layer_1_3_0(test_id: str, crs: str) -> QgsRasterLayer:
    basetestpath = tempfile.mkdtemp().replace("\\", "/")

    endpoint = basetestpath + f"/{test_id}_fake_qgis_http_endpoint"
    with open(
        QgisTestCase.sanitize_local_url(
            endpoint, "?SERVICE=WMS&REQUEST=GetCapabilities"
        ),
        "wb",
    ) as f:
        f.write(
            b"""
<WMS_Capabilities version="1.3.0" updateSequence="2872"
xmlns="http://www.opengis.net/wms"
xmlns:xlink="http://www.w3.org/1999/xlink"
xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.opengis.net/wms https://stadtplan.goettingen.de/geoserver/schemas/wms/1.3.0/capabilities_1_3_0.xsd">
<Service>
        <Name>WMS</Name>
        <Title>GeoServer Web Map Service</Title>
        <Abstract>A compliant implementation of WMS plus most of the SLD extension (dynamic styling). Can also generate PDF, SVG, KML, GeoRSS</Abstract>
</Service>
<Capability>
        <Request>
                <GetCapabilities>
                        <Format>text/xml</Format>
                </GetCapabilities>
                <GetMap>
                        <Format>image/png</Format>
                        <Format>image/jpeg</Format>
                        <Format>image/gif</Format>
                        <Format>image/GeoTIFF</Format>
                        <Format>image/tiff</Format>
                </GetMap>
                <GetFeatureInfo>
                        <Format>text/plain</Format>
                        <Format>text/html</Format>
                        <Format>text/xml</Format>
                </GetFeatureInfo>
        </Request>
        <Exception>
                <Format>XML</Format>
                <Format>INIMAGE</Format>
                <Format>BLANK</Format>
        </Exception>
        <Layer queryable="0" opaque="0" cascaded="1">
                <Name>de_basemapde_web_raster_farbe</Name>
                <Title>basemap.de Web Raster Farbe</Title>
                <Abstract>Der Layer enthaelt eine kombinierte Darstellung von ATKIS(c)- Landschaftsmodellen, der Hauskoordinaten und der Hausumringe,  die nach basemap.de WebSK in der jeweils aktuellen Version eine Kartensignatur haben, inklusive ihrer Beschriftung.</Abstract>
                <KeywordList/>
                <CRS>EPSG:3857</CRS>
                <CRS>CRS:84</CRS>
                <EX_GeographicBoundingBox>
                        <westBoundLongitude>0.1059467424056892</westBoundLongitude>
                        <eastBoundLongitude>20.448891294525627</eastBoundLongitude>
                        <southBoundLatitude>45.2375427360256</southBoundLatitude>
                        <northBoundLatitude>56.84787345153812</northBoundLatitude>
                </EX_GeographicBoundingBox>
                <BoundingBox CRS="CRS:84" minx="0.1059467424056892" miny="45.2375427360256" maxx="20.448891294525627" maxy="56.84787345153812"/>
                <BoundingBox CRS="EPSG:3857" minx="11793.937415807442" miny="5658995.571407319" maxx="2276360.1661935975" maxy="7729088.680470819"/>
                <Style>
                        <Name/>
                        <Title/>
                        <LegendURL width="2" height="1">
                                <Format>image/png</Format>
                                <OnlineResource
                                        xmlns:xlink="http://www.w3.org/1999/xlink" xlink:type="simple" xlink:href="https://stadtplan.goettingen.de/geoserver/goettingen/ows?service=WMS&amp;version=1.3.0&amp;request=GetLegendGraphic&amp;format=image%2Fpng&amp;width=20&amp;height=20&amp;layer=de_basemapde_web_raster_farbe"/>
                                </LegendURL>
                        </Style>
                </Layer>
        </Capability>
</WMS_Capabilities>"""
        )
    return QgsRasterLayer(
        f"crs={crs}&layers=de_basemapde_web_raster_farbe&styles&url=file://"
        + QgisTestCase.sanitize_local_url(
            endpoint, "?SERVICE=WMS&REQUEST=GetCapabilities"
        ),
        "test",
        "wms",
    )


def wms_layer_1_3_0_frankfurt(test_id: str, crs: str) -> QgsRasterLayer:
    basetestpath = tempfile.mkdtemp().replace("\\", "/")

    endpoint = basetestpath + f"/{test_id}_fake_qgis_http_endpoint"
    with open(
        QgisTestCase.sanitize_local_url(
            endpoint, "?SERVICE=WMS&REQUEST=GetCapabilities"
        ),
        "wb",
    ) as f:
        f.write(
            b"""
            <WMS_Capabilities xmlns="http://www.opengis.net/wms" xmlns:sld="http://www.opengis.net/sld" xmlns:xlink="http://www.w3.org/1999/xlink" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" version="1.3.0" xsi:schemaLocation="http://www.opengis.net/wms http://schemas.opengis.net/wms/1.3.0/capabilities_1_3_0.xsd http://www.opengis.net/sld http://schemas.opengis.net/sld/1.1.0/sld_capabilities.xsd">
            <Service>
              <Name>WMS</Name>
              <Title>MapProxy-WMS Rasterbilder Bebauungsplaene rv auf Basiskarte grau</Title>
              <Abstract>WMS mit Rasterbilder der rechtsverbindlichen Bebauungsplaene auf grauer Basiskarte ueber MapProxy</Abstract>
                <MaxWidth>4000</MaxWidth>
                <MaxHeight>4000</MaxHeight>
            </Service>
            <Capability>
              <Request>
                <GetCapabilities>
                  <Format>text/xml</Format>
                </GetCapabilities>
                <GetMap>
                  <Format>image/png</Format>
                  <Format>image/jpeg</Format>
                  <Format>image/gif</Format>
                  <Format>image/GeoTIFF</Format>
                  <Format>image/tiff</Format>
                </GetMap>
                <GetFeatureInfo>
                  <Format>text/plain</Format>
                  <Format>text/html</Format>
                  <Format>text/xml</Format>
                </GetFeatureInfo>
              </Request>
              <Exception>
                <Format>XML</Format>
                <Format>INIMAGE</Format>
                <Format>BLANK</Format>
              </Exception>
              <Layer>
                <Name>bplan_stadtkarte</Name>
                <Title>Bebauungsplaene Rasterbilder auf grauer Basiskarte</Title>
                <CRS>EPSG:25832</CRS>
                <CRS>EPSG:4326</CRS>
                <EX_GeographicBoundingBox>
                  <westBoundLongitude>8.471329688231897</westBoundLongitude>
                  <eastBoundLongitude>8.801042621684477</eastBoundLongitude>
                  <southBoundLatitude>50.01482739264088</southBoundLatitude>
                  <northBoundLatitude>50.22734893698391</northBoundLatitude>
                </EX_GeographicBoundingBox>
                <BoundingBox CRS="CRS:84" minx="8.471329688231897" miny="50.01482739264088" maxx="8.801042621684477" maxy="50.22734893698391" />
                <BoundingBox CRS="EPSG:4326" minx="50.01482739264088" miny="8.471329688231897" maxx="50.22734893698391" maxy="8.801042621684477" />
                <BoundingBox CRS="EPSG:25832" minx="462290" miny="5540412" maxx="485746" maxy="5563928" />
              </Layer>
            </Capability>
            </WMS_Capabilities>
            """
        )
    return QgsRasterLayer(
        f"crs={crs}&layers=bplan_stadtkarte&styles&url=file://"
        + QgisTestCase.sanitize_local_url(
            endpoint, "?SERVICE=WMS&REQUEST=GetCapabilities"
        ),
        "test",
        "wms",
    )



"""
***************************************************************************
    StatisticsByCategories.py
    ---------------------
    Date                 : September 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "September 2012"
__copyright__ = "(C) 2012, Victor Olaya"

from collections import defaultdict

from qgis.core import (
    NULL,
    QgsApplication,
    QgsCoordinateReferenceSystem,
    QgsDateTimeStatisticalSummary,
    QgsFeature,
    QgsFeatureRequest,
    QgsFeatureSink,
    QgsField,
    QgsFields,
    QgsProcessing,
    QgsProcessingException,
    QgsProcessingFeatureSource,
    QgsProcessingParameterFeatureSink,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterField,
    QgsStatisticalSummary,
    QgsStringStatisticalSummary,
    QgsWkbTypes,
)
from qgis.PyQt.QtCore import QMetaType

from processing.algs.qgis.QgisAlgorithm import QgisAlgorithm


class StatisticsByCategories(QgisAlgorithm):
    INPUT = "INPUT"
    VALUES_FIELD_NAME = "VALUES_FIELD_NAME"
    CATEGORIES_FIELD_NAME = "CATEGORIES_FIELD_NAME"
    OUTPUT = "OUTPUT"

    def group(self):
        return self.tr("Vector analysis")

    def groupId(self):
        return "vectoranalysis"

    def tags(self):
        return self.tr(
            "groups,stats,statistics,table,layer,sum,maximum,minimum,mean,average,standard,deviation,"
            "count,distinct,unique,variance,median,quartile,range,majority,minority,histogram,distinct,summary"
        ).split(",")

    def icon(self):
        return QgsApplication.getThemeIcon("/algorithms/mAlgorithmBasicStatistics.svg")

    def svgIconPath(self):
        return QgsApplication.iconPath("/algorithms/mAlgorithmBasicStatistics.svg")

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterFeatureSource(
                self.INPUT,
                self.tr("Input vector layer"),
                types=[QgsProcessing.SourceType.TypeVector],
            )
        )
        self.addParameter(
            QgsProcessingParameterField(
                self.VALUES_FIELD_NAME,
                self.tr(
                    "Field to calculate statistics on (if empty, only count is calculated)"
                ),
                parentLayerParameterName=self.INPUT,
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterField(
                self.CATEGORIES_FIELD_NAME,
                self.tr("Field(s) with categories"),
                parentLayerParameterName=self.INPUT,
                type=QgsProcessingParameterField.DataType.Any,
                allowMultiple=True,
            )
        )

        self.addParameter(
            QgsProcessingParameterFeatureSink(
                self.OUTPUT, self.tr("Statistics by category")
            )
        )

    def name(self):
        return "statisticsbycategories"

    def displayName(self):
        return self.tr("Statistics by categories")

    def shortDescription(self):
        return self.tr("Calculates statistics of fields depending on a parent class.")

    def shortHelpString(self):
        return self.tr(
            "This algorithm calculates statistics of fields depending on a parent class."
        )

    def processAlgorithm(self, parameters, context, feedback):
        source = self.parameterAsSource(parameters, self.INPUT, context)
        if source is None:
            raise QgsProcessingException(
                self.invalidSourceError(parameters, self.INPUT)
            )

        value_field_name = self.parameterAsString(
            parameters, self.VALUES_FIELD_NAME, context
        )
        category_field_names = self.parameterAsFields(
            parameters, self.CATEGORIES_FIELD_NAME, context
        )

        value_field_index = source.fields().lookupField(value_field_name)
        if value_field_index >= 0:
            value_field = source.fields().at(value_field_index)
        else:
            value_field = None
        category_field_indexes = list()

        # generate output fields
        fields = QgsFields()
        for field_name in category_field_names:
            c = source.fields().lookupField(field_name)
            if c == -1:
                raise QgsProcessingException(
                    self.tr('Field "{field_name}" does not exist.').format(
                        field_name=field_name
                    )
                )
            category_field_indexes.append(c)
            fields.append(source.fields().at(c))

        def addField(name):
            """
            Adds a field to the output, keeping the same data type as the value_field
            """
            field = QgsField(value_field)
            field.setName(name)
            fields.append(field)

        if value_field is None:
            field_type = "none"
            fields.append(QgsField("count", QMetaType.Type.Int))
        elif value_field.isNumeric():
            field_type = "numeric"
            fields.append(QgsField("count", QMetaType.Type.Int))
            fields.append(QgsField("unique", QMetaType.Type.Int))
            fields.append(QgsField("min", QMetaType.Type.Double))
            fields.append(QgsField("max", QMetaType.Type.Double))
            fields.append(QgsField("range", QMetaType.Type.Double))
            fields.append(QgsField("sum", QMetaType.Type.Double))
            fields.append(QgsField("mean", QMetaType.Type.Double))
            fields.append(QgsField("median", QMetaType.Type.Double))
            fields.append(QgsField("stddev", QMetaType.Type.Double))
            fields.append(QgsField("minority", QMetaType.Type.Double))
            fields.append(QgsField("majority", QMetaType.Type.Double))
            fields.append(QgsField("q1", QMetaType.Type.Double))
            fields.append(QgsField("q3", QMetaType.Type.Double))
            fields.append(QgsField("iqr", QMetaType.Type.Double))
        elif value_field.type() in (
            QMetaType.Type.QDate,
            QMetaType.Type.QTime,
            QMetaType.Type.QDateTime,
        ):
            field_type = "datetime"
            fields.append(QgsField("count", QMetaType.Type.Int))
            fields.append(QgsField("unique", QMetaType.Type.Int))
            fields.append(QgsField("empty", QMetaType.Type.Int))
            fields.append(QgsField("filled", QMetaType.Type.Int))
            # keep same data type for these fields
            addField("min")
            addField("max")
        else:
            field_type = "string"
            fields.append(QgsField("count", QMetaType.Type.Int))
            fields.append(QgsField("unique", QMetaType.Type.Int))
            fields.append(QgsField("empty", QMetaType.Type.Int))
            fields.append(QgsField("filled", QMetaType.Type.Int))
            # keep same data type for these fields
            addField("min")
            addField("max")
            fields.append(QgsField("min_length", QMetaType.Type.Int))
            fields.append(QgsField("max_length", QMetaType.Type.Int))
            fields.append(QgsField("mean_length", QMetaType.Type.Double))

        request = QgsFeatureRequest().setFlags(QgsFeatureRequest.Flag.NoGeometry)
        if value_field is not None:
            attrs = [value_field_index]
        else:
            attrs = []
        attrs.extend(category_field_indexes)
        request.setSubsetOfAttributes(attrs)
        features = source.getFeatures(
            request, QgsProcessingFeatureSource.Flag.FlagSkipGeometryValidityChecks
        )
        total = 50.0 / source.featureCount() if source.featureCount() else 0
        if field_type == "none":
            values = defaultdict(int)
        else:
            values = defaultdict(list)
        for current, feat in enumerate(features):
            if feedback.isCanceled():
                break

            feedback.setProgress(int(current * total))
            attrs = feat.attributes()
            cat = tuple([attrs[c] for c in category_field_indexes])
            if field_type == "none":
                values[cat] += 1
                continue
            if field_type == "numeric":
                if attrs[value_field_index] == NULL:
                    continue
                else:
                    value = float(attrs[value_field_index])
            elif field_type == "string":
                if attrs[value_field_index] == NULL:
                    value = ""
                else:
                    value = str(attrs[value_field_index])
            elif attrs[value_field_index] == NULL:
                value = NULL
            else:
                value = attrs[value_field_index]
            values[cat].append(value)

        (sink, dest_id) = self.parameterAsSink(
            parameters,
            self.OUTPUT,
            context,
            fields,
            QgsWkbTypes.Type.NoGeometry,
            QgsCoordinateReferenceSystem(),
        )
        if sink is None:
            raise QgsProcessingException(self.invalidSinkError(parameters, self.OUTPUT))

        if field_type == "none":
            self.saveCounts(values, sink, feedback)
        elif field_type == "numeric":
            self.calcNumericStats(values, sink, feedback)
        elif field_type == "datetime":
            self.calcDateTimeStats(values, sink, feedback)
        else:
            self.calcStringStats(values, sink, feedback)

        sink.finalize()
        feedback.featureSinkFinalized(self.OUTPUT)
        return {self.OUTPUT: dest_id}

    def saveCounts(self, values, sink, feedback):
        total = 50.0 / len(values) if values else 0
        current = 0
        for cat, v in values.items():
            if feedback.isCanceled():
                break

            feedback.setProgress(int(current * total) + 50)
            f = QgsFeature()
            f.setAttributes(list(cat) + [v])
            sink.addFeature(f, QgsFeatureSink.Flag.FastInsert)
            current += 1

    def calcNumericStats(self, values, sink, feedback):
        stat = QgsStatisticalSummary()

        total = 50.0 / len(values) if values else 0
        current = 0
        for cat, v in values.items():
            if feedback.isCanceled():
                break

            feedback.setProgress(int(current * total) + 50)

            stat.calculate(v)
            f = QgsFeature()
            f.setAttributes(
                list(cat)
                + [
                    stat.count(),
                    stat.variety(),
                    stat.min(),
                    stat.max(),
                    stat.range(),
                    stat.sum(),
                    stat.mean(),
                    stat.median(),
                    stat.stDev(),
                    stat.minority(),
                    stat.majority(),
                    stat.firstQuartile(),
                    stat.thirdQuartile(),
                    stat.interQuartileRange(),
                ]
            )

            sink.addFeature(f, QgsFeatureSink.Flag.FastInsert)
            current += 1

    def calcDateTimeStats(self, values, sink, feedback):
        stat = QgsDateTimeStatisticalSummary()

        total = 50.0 / len(values) if values else 0
        current = 0
        for cat, v in values.items():
            if feedback.isCanceled():
                break

            feedback.setProgress(int(current * total) + 50)

            stat.calculate(v)
            f = QgsFeature()
            f.setAttributes(
                list(cat)
                + [
                    stat.count(),
                    stat.countDistinct(),
                    stat.countMissing(),
                    stat.count() - stat.countMissing(),
                    stat.statistic(QgsDateTimeStatisticalSummary.Statistic.Min),
                    stat.statistic(QgsDateTimeStatisticalSummary.Statistic.Max),
                ]
            )

            sink.addFeature(f, QgsFeatureSink.Flag.FastInsert)
            current += 1

    def calcStringStats(self, values, sink, feedback):
        stat = QgsStringStatisticalSummary()

        total = 50.0 / len(values) if values else 0
        current = 0
        for cat, v in values.items():
            if feedback.isCanceled():
                break

            feedback.setProgress(int(current * total) + 50)

            stat.calculate(v)
            f = QgsFeature()
            f.setAttributes(
                list(cat)
                + [
                    stat.count(),
                    stat.countDistinct(),
                    stat.countMissing(),
                    stat.count() - stat.countMissing(),
                    stat.min(),
                    stat.max(),
                    stat.minLength(),
                    stat.maxLength(),
                    stat.meanLength(),
                ]
            )

            sink.addFeature(f, QgsFeatureSink.Flag.FastInsert)
            current += 1



"""
***************************************************************************
    general.py
    ---------------------
    Date                 : April 2013
    Copyright            : (C) 2013 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "April 2013"
__copyright__ = "(C) 2013, Victor Olaya"

from qgis.core import (
    QgsApplication,
    QgsProcessingAlgorithm,
    QgsProcessingOutputLayerDefinition,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterEnum,
    QgsProcessingParameterFeatureSink,
    QgsProcessingParameterRasterDestination,
    QgsProcessingParameterVectorDestination,
    QgsProject,
)
from qgis.utils import iface

from processing.core.Processing import Processing
from processing.gui.algorithm_widget import AlgorithmWidget
from processing.gui.Postprocessing import handleAlgorithmResults


# changing this signature? make sure you update the signature in
# python/processing/__init__.py too!
# Docstring for this function is in python/processing/__init__.py
def algorithmHelp(id: str) -> None:
    alg = QgsApplication.processingRegistry().algorithmById(id)
    if alg is not None:
        print(f"{alg.displayName()} ({alg.id()})\n")
        if alg.shortDescription():
            print(alg.shortDescription() + "\n")
        if alg.shortHelpString():
            print(alg.shortHelpString() + "\n")
        print("\n----------------")
        print("Input parameters")
        print("----------------")
        for p in alg.parameterDefinitions():
            if p.flags() & QgsProcessingParameterDefinition.Flag.FlagHidden:
                continue
            print(f"\n{p.name()}: {p.description()}")
            if p.help():
                print(f"\n\t{p.help()}")

            print(f"\n\tParameter type:\t{p.__class__.__name__}")

            if isinstance(p, QgsProcessingParameterEnum):
                opts = []
                for i, o in enumerate(p.options()):
                    opts.append(f"\t\t- {i}: {o}")
                print("\n\tAvailable values:\n{}".format("\n".join(opts)))

            parameter_type = QgsApplication.processingRegistry().parameterType(p.type())
            accepted_types = (
                parameter_type.acceptedPythonTypes()
                if parameter_type is not None
                else []
            )
            if accepted_types:
                opts = []
                for t in accepted_types:
                    opts.append(f"\t\t- {t}")
                print("\n\tAccepted data types:")
                print("\n".join(opts))

        print("\n----------------")
        print("Outputs")
        print("----------------")

        for o in alg.outputDefinitions():
            print(f"\n{o.name()}:  <{o.__class__.__name__}>")
            if o.description():
                print("\t" + o.description())

    else:
        print(f'Algorithm "{id}" not found.')


# changing this signature? make sure you update the signature in
# python/processing/__init__.py too!
# Docstring for this function is in python/processing/__init__.py
def run(
    algOrName,
    parameters,
    onFinish=None,
    feedback=None,
    context=None,
    is_child_algorithm=False,
):
    if onFinish or not is_child_algorithm:
        return Processing.runAlgorithm(
            algOrName, parameters, onFinish, feedback, context
        )
    else:
        # for child algorithms, we disable to default post-processing step where layer ownership
        # is transferred from the context to the caller. In this case, we NEED the ownership to remain
        # with the context, so that further steps in the algorithm have guaranteed access to the layer.
        def post_process(_alg, _context, _feedback):
            return

        return Processing.runAlgorithm(
            algOrName,
            parameters,
            onFinish=post_process,
            feedback=feedback,
            context=context,
        )


# changing this signature? make sure you update the signature in
# python/processing/__init__.py too!
# Docstring for this function is in python/processing/__init__.py
def runAndLoadResults(algOrName, parameters, feedback=None, context=None):
    if isinstance(algOrName, QgsProcessingAlgorithm):
        alg = algOrName
    else:
        alg = QgsApplication.processingRegistry().createAlgorithmById(algOrName)

    # output destination parameters to point to current project
    for param in alg.parameterDefinitions():
        if not param.name() in parameters:
            continue

        if isinstance(
            param,
            (
                QgsProcessingParameterFeatureSink,
                QgsProcessingParameterVectorDestination,
                QgsProcessingParameterRasterDestination,
            ),
        ):
            p = parameters[param.name()]
            if not isinstance(p, QgsProcessingOutputLayerDefinition):
                parameters[param.name()] = QgsProcessingOutputLayerDefinition(
                    p, QgsProject.instance()
                )
            else:
                p.destinationProject = QgsProject.instance()
                parameters[param.name()] = p

    return Processing.runAlgorithm(
        alg,
        parameters=parameters,
        onFinish=handleAlgorithmResults,
        feedback=feedback,
        context=context,
    )


# changing this signature? make sure you update the signature in
# python/processing/__init__.py too!
# Docstring for this function is in python/processing/__init__.py
def createAlgorithmDialog(algOrName, parameters={}):
    if isinstance(algOrName, QgsProcessingAlgorithm):
        alg = algOrName.create()
    else:
        alg = QgsApplication.processingRegistry().createAlgorithmById(algOrName)

    if alg is None:
        return None

    widget = alg.createCustomParametersWidget(iface.mainWindow())

    if not widget:
        widget = AlgorithmWidget(alg, parent=iface.mainWindow())

    widget.setParameters(parameters)

    return widget


# changing this signature? make sure you update the signature in
# python/processing/__init__.py too!
# Docstring for this function is in python/processing/__init__.py
def execAlgorithmDialog(algOrName, parameters={}):
    widget = createAlgorithmDialog(algOrName, parameters)
    if widget is None:
        return {}

    widget.exec()

    results = widget.results()
    # make sure the dialog is destroyed and not only hidden on pressing Esc
    widget.close()
    return results



"""
***************************************************************************
    ConfigDialog.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import os
import warnings

from qgis.core import NULL, QgsApplication, QgsSettings
from qgis.gui import (
    QgsDoubleSpinBox,
    QgsOptionsDialogHighlightWidget,
    QgsOptionsPageWidget,
    QgsSpinBox,
)
from qgis.PyQt import uic
from qgis.PyQt.QtCore import QEvent, Qt
from qgis.PyQt.QtGui import QCursor, QIcon, QStandardItem, QStandardItemModel
from qgis.PyQt.QtWidgets import (
    QApplication,
    QComboBox,
    QFileDialog,
    QHBoxLayout,
    QLineEdit,
    QMessageBox,
    QPushButton,
    QStyle,
    QStyledItemDelegate,
    QToolButton,
    QWidget,
)
from qgis.utils import OverrideCursor

from processing.core.Processing import Processing
from processing.core.ProcessingConfig import ProcessingConfig, Setting, settingsWatcher
from processing.gui.DirectorySelectorDialog import DirectorySelectorDialog
from processing.gui.menus import defaultMenuEntries, menusSettingsGroup

pluginPath = os.path.split(os.path.dirname(__file__))[0]
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    WIDGET, BASE = uic.loadUiType(os.path.join(pluginPath, "ui", "DlgConfig.ui"))


class ConfigOptionsPage(QgsOptionsPageWidget):
    def __init__(self, parent):
        super().__init__(parent)
        self.config_widget = ConfigDialog(False)
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setMargin(0)
        self.setLayout(layout)
        layout.addWidget(self.config_widget)
        self.setObjectName("processingOptions")
        self.highlightWidget = ProcessingTreeHighlight(self.config_widget)
        self.registerHighlightWidget(self.highlightWidget)

    def apply(self):
        self.config_widget.accept()

    def helpKey(self):
        return "processing/index.html"


class ProcessingTreeHighlight(QgsOptionsDialogHighlightWidget):
    def __init__(self, config_dialog):
        super().__init__(config_dialog.tree)
        self.config_dialog = config_dialog

    def highlightText(self, text):
        return self.config_dialog.textChanged(text)

    def searchText(self, text):
        return self.config_dialog.textChanged(text)

    def reset(self):
        self.config_dialog.textChanged("")


class ConfigDialog(BASE, WIDGET):
    def __init__(self, showSearch=True):
        super().__init__(None)
        self.setupUi(self)

        self.groupIcon = QgsApplication.getThemeIcon("mIconFolder.svg")

        self.model = QStandardItemModel()
        self.tree.setModel(self.model)

        self.delegate = SettingDelegate()
        self.tree.setItemDelegateForColumn(1, self.delegate)

        if showSearch:
            if hasattr(self.searchBox, "setPlaceholderText"):
                self.searchBox.setPlaceholderText(
                    QApplication.translate("ConfigDialog", "Search…")
                )
            self.searchBox.textChanged.connect(self.textChanged)
        else:
            self.searchBox.hide()

        self.fillTree()

        self.saveMenus = False
        self.tree.expanded.connect(self.itemExpanded)
        self.auto_adjust_columns = True

    def textChanged(self, text=None):
        if text is not None:
            text = str(text.lower())
        else:
            text = str(self.searchBox.text().lower())
        found = self._filterItem(
            self.model.invisibleRootItem(), text, False if text else True
        )

        self.auto_adjust_columns = False
        if text:
            self.tree.expandAll()
        else:
            self.tree.collapseAll()
        self.adjustColumns()
        self.auto_adjust_columns = True

        if text:
            return found
        else:
            self.tree.collapseAll()
            return False

    def _filterItem(self, item, text, forceShow=False):
        if item.hasChildren():
            show = (
                forceShow
                or isinstance(item, QStandardItem)
                and bool(text)
                and (text in item.text().lower())
            )
            for i in range(item.rowCount()):
                child = item.child(i)
                show = self._filterItem(child, text, forceShow) or show
            self.tree.setRowHidden(item.row(), item.index().parent(), not show)
            return show

        elif isinstance(item, QStandardItem):
            show = forceShow or bool(text) and (text in item.text().lower())
            self.tree.setRowHidden(item.row(), item.index().parent(), not show)
            return show

    def fillTree(self):
        self.fillTreeUsingProviders()

    def fillTreeUsingProviders(self):
        self.items = {}
        self.model.clear()
        self.model.setHorizontalHeaderLabels([self.tr("Setting"), self.tr("Value")])

        settings = ProcessingConfig.getSettings()

        rootItem = self.model.invisibleRootItem()

        """
        Filter 'General', 'Models' and 'Scripts' items
        """
        priorityKeys = [self.tr("General"), self.tr("Models"), self.tr("Scripts")]
        for group in priorityKeys:
            groupItem = QStandardItem(group)
            icon = ProcessingConfig.getGroupIcon(group)
            groupItem.setIcon(icon)
            groupItem.setEditable(False)
            emptyItem = QStandardItem()
            emptyItem.setEditable(False)

            rootItem.insertRow(0, [groupItem, emptyItem])
            if group not in settings:
                continue

            # add menu item only if it has any search matches
            for setting in settings[group]:
                if setting.hidden or setting.name.startswith("MENU_"):
                    continue

                labelItem = QStandardItem(setting.description)
                labelItem.setIcon(icon)
                labelItem.setEditable(False)
                self.items[setting] = SettingItem(setting)
                groupItem.insertRow(0, [labelItem, self.items[setting]])

        """
        Filter 'Providers' items
        """
        providersItem = QStandardItem(self.tr("Providers"))
        icon = QgsApplication.getThemeIcon("/processingAlgorithm.svg")
        providersItem.setIcon(icon)
        providersItem.setEditable(False)
        emptyItem = QStandardItem()
        emptyItem.setEditable(False)

        rootItem.insertRow(0, [providersItem, emptyItem])
        for group in list(settings.keys()):
            if group in priorityKeys or group == menusSettingsGroup:
                continue

            groupItem = QStandardItem(group)
            icon = ProcessingConfig.getGroupIcon(group)
            groupItem.setIcon(icon)
            groupItem.setEditable(False)

            for setting in settings[group]:
                if setting.hidden:
                    continue

                labelItem = QStandardItem(setting.description)
                labelItem.setIcon(icon)
                labelItem.setEditable(False)
                self.items[setting] = SettingItem(setting)
                groupItem.insertRow(0, [labelItem, self.items[setting]])

            emptyItem = QStandardItem()
            emptyItem.setEditable(False)
            providersItem.appendRow([groupItem, emptyItem])

        """
        Filter 'Menus' items
        """
        self.menusItem = QStandardItem(self.tr("Menus"))
        icon = QIcon(os.path.join(pluginPath, "images", "menu.png"))
        self.menusItem.setIcon(icon)
        self.menusItem.setEditable(False)
        emptyItem = QStandardItem()
        emptyItem.setEditable(False)

        rootItem.insertRow(0, [self.menusItem, emptyItem])

        button = QPushButton(self.tr("Reset to defaults"))
        button.clicked.connect(self.resetMenusToDefaults)
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(button)
        layout.addStretch()
        widget = QWidget()
        widget.setLayout(layout)
        self.tree.setIndexWidget(emptyItem.index(), widget)

        for provider in QgsApplication.processingRegistry().providers():
            providerDescription = provider.name()
            groupItem = QStandardItem(providerDescription)
            icon = provider.icon()
            groupItem.setIcon(icon)
            groupItem.setEditable(False)

            for alg in provider.algorithms():
                algItem = QStandardItem(alg.displayName())
                algItem.setIcon(icon)
                algItem.setEditable(False)
                try:
                    settingMenu = ProcessingConfig.settings["MENU_" + alg.id()]
                    settingButton = ProcessingConfig.settings["BUTTON_" + alg.id()]
                    settingIcon = ProcessingConfig.settings["ICON_" + alg.id()]
                except:
                    continue
                self.items[settingMenu] = SettingItem(settingMenu)
                self.items[settingButton] = SettingItem(settingButton)
                self.items[settingIcon] = SettingItem(settingIcon)
                menuLabelItem = QStandardItem("Menu path")
                menuLabelItem.setEditable(False)
                buttonLabelItem = QStandardItem("Add button in toolbar")
                buttonLabelItem.setEditable(False)
                iconLabelItem = QStandardItem("Icon")
                iconLabelItem.setEditable(False)
                emptyItem = QStandardItem()
                emptyItem.setEditable(False)
                algItem.insertRow(0, [menuLabelItem, self.items[settingMenu]])
                algItem.insertRow(0, [buttonLabelItem, self.items[settingButton]])
                algItem.insertRow(0, [iconLabelItem, self.items[settingIcon]])
                groupItem.insertRow(0, [algItem, emptyItem])

            emptyItem = QStandardItem()
            emptyItem.setEditable(False)

            self.menusItem.appendRow([groupItem, emptyItem])

        self.tree.sortByColumn(0, Qt.SortOrder.AscendingOrder)
        self.adjustColumns()

    def resetMenusToDefaults(self):
        for provider in QgsApplication.processingRegistry().providers():
            for alg in provider.algorithms():
                d = defaultMenuEntries.get(alg.id(), "")
                setting = ProcessingConfig.settings["MENU_" + alg.id()]
                item = self.items[setting]
                item.setData(d, Qt.ItemDataRole.EditRole)
        self.saveMenus = True

    def accept(self):
        qsettings = QgsSettings()
        for setting in list(self.items.keys()):
            if setting.group != menusSettingsGroup or self.saveMenus:
                if isinstance(setting.value, bool):
                    setting.setValue(
                        self.items[setting].checkState() == Qt.CheckState.Checked
                    )
                else:
                    try:
                        setting.setValue(str(self.items[setting].text()))
                    except ValueError as e:
                        QMessageBox.warning(
                            self,
                            self.tr("Wrong value"),
                            self.tr('Wrong value for parameter "{0}":\n\n{1}').format(
                                setting.description, str(e)
                            ),
                        )
                        return
                setting.save(qsettings)

        with OverrideCursor(Qt.CursorShape.WaitCursor):
            for p in QgsApplication.processingRegistry().providers():
                p.refreshAlgorithms()

        settingsWatcher.settingsChanged.emit()

    def itemExpanded(self, idx):
        if idx == self.menusItem.index():
            self.saveMenus = True
        if self.auto_adjust_columns:
            self.adjustColumns()

    def adjustColumns(self):
        self.tree.resizeColumnToContents(0)
        self.tree.resizeColumnToContents(1)


class SettingItem(QStandardItem):
    def __init__(self, setting):
        QStandardItem.__init__(self)
        self.setting = setting
        self.setData(setting, Qt.ItemDataRole.UserRole)
        if isinstance(setting.value, bool):
            self.setCheckable(True)
            self.setEditable(False)
            if setting.value:
                self.setCheckState(Qt.CheckState.Checked)
            else:
                self.setCheckState(Qt.CheckState.Unchecked)
        else:
            self.setData(setting.value, Qt.ItemDataRole.EditRole)


class SettingDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def createEditor(self, parent, options, index):
        setting = index.model().data(index, Qt.ItemDataRole.UserRole)
        if setting.valuetype == Setting.FOLDER:
            return FileDirectorySelector(parent, placeholder=setting.placeholder)
        elif setting.valuetype == Setting.FILE:
            return FileDirectorySelector(parent, True, setting.placeholder)
        elif setting.valuetype == Setting.SELECTION:
            combo = QComboBox(parent)
            combo.addItems(setting.options)
            return combo
        elif setting.valuetype == Setting.SELECTION_STORE_STRING:
            combo = QComboBox(parent)
            for option in setting.options:
                combo.addItem(option, option)
            return combo
        elif setting.valuetype == Setting.MULTIPLE_FOLDERS:
            return MultipleDirectorySelector(parent, setting.placeholder)
        else:
            value = self.convertValue(
                index.model().data(index, Qt.ItemDataRole.EditRole)
            )
            if isinstance(value, int):
                spnBox = QgsSpinBox(parent)
                spnBox.setRange(-999999999, 999999999)
                return spnBox
            elif isinstance(value, float):
                spnBox = QgsDoubleSpinBox(parent)
                spnBox.setRange(-999999999.999999, 999999999.999999)
                spnBox.setDecimals(6)
                return spnBox
            elif isinstance(value, str):
                lineEdit = QLineEdit(parent)
                lineEdit.setPlaceholderText(setting.placeholder)
                return lineEdit

    def setEditorData(self, editor, index):
        value = self.convertValue(index.model().data(index, Qt.ItemDataRole.EditRole))
        setting = index.model().data(index, Qt.ItemDataRole.UserRole)
        if setting.valuetype == Setting.SELECTION:
            editor.setCurrentIndex(editor.findText(value))
        elif setting.valuetype == Setting.SELECTION_STORE_STRING:
            editor.setCurrentIndex(editor.findData(value))
        elif setting.valuetype in (Setting.FLOAT, Setting.INT):
            editor.setValue(value)
        else:
            editor.setText(value)

    def setModelData(self, editor, model, index):
        value = self.convertValue(index.model().data(index, Qt.ItemDataRole.EditRole))
        setting = index.model().data(index, Qt.ItemDataRole.UserRole)
        if setting.valuetype == Setting.SELECTION:
            model.setData(index, editor.currentText(), Qt.ItemDataRole.EditRole)
        elif setting.valuetype == Setting.SELECTION_STORE_STRING:
            model.setData(index, editor.currentData(), Qt.ItemDataRole.EditRole)
        else:
            if isinstance(value, str):
                model.setData(index, editor.text(), Qt.ItemDataRole.EditRole)
            else:
                model.setData(index, editor.value(), Qt.ItemDataRole.EditRole)

    def sizeHint(self, option, index):
        return QgsSpinBox().sizeHint()

    def eventFilter(self, editor, event):
        if event.type() == QEvent.Type.FocusOut and hasattr(editor, "canFocusOut"):
            if not editor.canFocusOut:
                return False
        return QStyledItemDelegate.eventFilter(self, editor, event)

    def convertValue(self, value):
        if value is None or value == NULL:
            return ""
        try:
            return int(value)
        except:
            try:
                return float(value)
            except:
                return str(value)


class FileDirectorySelector(QWidget):
    def __init__(self, parent=None, selectFile=False, placeholder=""):
        QWidget.__init__(self, parent)

        # create gui
        self.btnSelect = QToolButton()
        self.btnSelect.setText("…")
        self.lineEdit = QLineEdit()
        self.lineEdit.setPlaceholderText(placeholder)
        self.hbl = QHBoxLayout()
        self.hbl.setMargin(0)
        self.hbl.setSpacing(0)
        self.hbl.addWidget(self.lineEdit)
        self.hbl.addWidget(self.btnSelect)

        self.setLayout(self.hbl)

        self.canFocusOut = False
        self.selectFile = selectFile

        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        self.btnSelect.clicked.connect(self.select)

    def select(self):
        lastDir = ""
        if not self.selectFile:
            selectedPath = QFileDialog.getExistingDirectory(
                None,
                self.tr("Select directory"),
                lastDir,
                QFileDialog.Option.ShowDirsOnly,
            )
        else:
            selectedPath, selected_filter = QFileDialog.getOpenFileName(
                None, self.tr("Select file"), lastDir, self.tr("All files (*)")
            )

        if not selectedPath:
            return

        self.lineEdit.setText(selectedPath)
        self.canFocusOut = True

    def text(self):
        return self.lineEdit.text()

    def setText(self, value):
        self.lineEdit.setText(value)


class MultipleDirectorySelector(QWidget):
    def __init__(self, parent=None, placeholder=""):
        QWidget.__init__(self, parent)

        # create gui
        self.btnSelect = QToolButton()
        self.btnSelect.setText("…")
        self.lineEdit = QLineEdit()
        self.lineEdit.setPlaceholderText(placeholder)
        self.hbl = QHBoxLayout()
        self.hbl.setMargin(0)
        self.hbl.setSpacing(0)
        self.hbl.addWidget(self.lineEdit)
        self.hbl.addWidget(self.btnSelect)

        self.setLayout(self.hbl)

        self.canFocusOut = False

        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        self.btnSelect.clicked.connect(self.select)

    def select(self):
        text = self.lineEdit.text()
        if text != "":
            items = text.split(";")
        else:
            items = []

        dlg = DirectorySelectorDialog(None, items)
        if dlg.exec():
            text = dlg.value()
            self.lineEdit.setText(text)

        self.canFocusOut = True

    def text(self):
        return self.lineEdit.text()

    def setText(self, value):
        self.lineEdit.setText(value)



"""
***************************************************************************
    GdalAlgorithmDialog.py
    ---------------------
    Date                 : May 2015
    Copyright            : (C) 2015 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "May 2015"
__copyright__ = "(C) 2015, Victor Olaya"

from qgis.core import (
    QgsProcessingException,
    QgsProcessingFeedback,
    QgsProcessingParameterDefinition,
)
from qgis.gui import (
    QgsMessageBar,
    QgsProcessingLayerOutputDestinationWidget,
    QgsProjectionSelectionWidget,
)
from qgis.PyQt.QtCore import QCoreApplication
from qgis.PyQt.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialogButtonBox,
    QLabel,
    QLineEdit,
    QPlainTextEdit,
    QPushButton,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from processing.core.exceptions import InvalidOutputExtension, InvalidParameterValue
from processing.gui.algorithm_widget import AlgorithmWidget
from processing.gui.MultipleInputPanel import MultipleInputPanel
from processing.gui.NumberInputPanel import NumberInputPanel
from processing.gui.ParametersPanel import ParametersPanel
from processing.tools.dataobjects import createContext


class GdalAlgorithmWidget(AlgorithmWidget):
    """
    Custom algorithm widget for showing GDAL command line arguments
    """

    def __init__(self, alg, parent=None):
        super().__init__(alg, parent=parent)
        self.mainWidget().parametersHaveChanged()

    def getParametersPanel(self, alg, parent):
        return GdalParametersPanel(parent, alg)


class GdalParametersPanel(ParametersPanel):
    def __init__(self, parent, alg):
        super().__init__(parent, alg)

        self.dialog = parent
        w = QWidget()
        layout = QVBoxLayout()
        layout.setMargin(0)
        layout.setSpacing(6)
        label = QLabel()
        label.setText(self.tr("GDAL/OGR console call"))
        layout.addWidget(label)
        self.text = QPlainTextEdit()
        self.text.setReadOnly(True)
        layout.addWidget(self.text)
        w.setLayout(layout)
        self.addExtraWidget(w)

        self.connectParameterSignals()
        self.parametersHaveChanged()

    def connectParameterSignals(self):
        for wrapper in list(self.wrappers.values()):
            wrapper.widgetValueHasChanged.connect(self.parametersHaveChanged)

            # TODO - remove when all wrappers correctly emit widgetValueHasChanged!
            w = wrapper.wrappedWidget()
            self.connectWidgetChangedSignals(w)
            for c in w.findChildren(QWidget):
                self.connectWidgetChangedSignals(c)

    def connectWidgetChangedSignals(self, w):
        if isinstance(w, QLineEdit):
            w.textChanged.connect(self.parametersHaveChanged)
        elif isinstance(w, QComboBox):
            w.currentIndexChanged.connect(self.parametersHaveChanged)
        elif isinstance(w, QgsProjectionSelectionWidget):
            w.crsChanged.connect(self.parametersHaveChanged)
        elif isinstance(w, QCheckBox):
            w.stateChanged.connect(self.parametersHaveChanged)
        elif isinstance(w, MultipleInputPanel):
            w.selectionChanged.connect(self.parametersHaveChanged)
        elif isinstance(w, NumberInputPanel):
            w.hasChanged.connect(self.parametersHaveChanged)
        elif isinstance(w, QgsProcessingLayerOutputDestinationWidget):
            w.destinationChanged.connect(self.parametersHaveChanged)

    def parametersHaveChanged(self):
        context = createContext()
        feedback = QgsProcessingFeedback()
        try:
            # messy as all heck, but we don't want to call the dialog's implementation of
            # createProcessingParameters as we want to catch the exceptions raised by the
            # parameter panel instead...
            parameters = (
                {}
                if self.dialog.mainWidget() is None
                else self.dialog.mainWidget().createProcessingParameters()
            )
            for output in self.algorithm().destinationParameterDefinitions():
                if not output.name() in parameters or parameters[output.name()] is None:
                    if (
                        not output.flags()
                        & QgsProcessingParameterDefinition.Flag.FlagOptional
                    ):
                        parameters[output.name()] = self.tr("[temporary file]")
            for p in self.algorithm().parameterDefinitions():
                if p.flags() & QgsProcessingParameterDefinition.Flag.FlagHidden:
                    continue

                if (
                    p.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
                    and p.name() not in parameters
                ):
                    continue

                if p.name() not in parameters or not p.checkValueIsAcceptable(
                    parameters[p.name()]
                ):
                    # not ready yet
                    self.text.setPlainText("")
                    return

            try:
                commands = self.algorithm().getConsoleCommands(
                    parameters, context, feedback, executing=False
                )
                commands = [c for c in commands if c not in ["cmd.exe", "/C "]]
                self.text.setPlainText(" ".join(commands))
            except QgsProcessingException as e:
                self.text.setPlainText(str(e))
        except InvalidParameterValue as e:
            self.text.setPlainText(
                self.tr("Invalid value for parameter '{0}'").format(
                    e.parameter.description()
                )
            )
        except InvalidOutputExtension as e:
            self.text.setPlainText(e.message)
